from typing import TYPE_CHECKING, ItemsView, List, Optional, Union, ValuesView

from pydantic import BaseModel, Field, ValidationInfo, field_validator

from trilogy.constants import DEFAULT_NAMESPACE, logger
from trilogy.core.enums import Modifier
from trilogy.core.models.author import (
    Concept,
    ConceptRef,
    Function,
    Grain,
    HasUUID,
    LooseConceptList,
    Namespaced,
    WhereClause,
)

LOGGER_PREFIX = "[MODELS_DATASOURCE]"

if TYPE_CHECKING:
    pass


class RawColumnExpr(BaseModel):
    text: str


class ColumnAssignment(BaseModel):
    alias: str | RawColumnExpr | Function
    concept: ConceptRef
    modifiers: List[Modifier] = Field(default_factory=list)

    @field_validator("concept", mode="before")
    def force_reference(cls, v: ConceptRef, info: ValidationInfo):
        if isinstance(v, Concept):
            return v.reference
        return v

    def __eq__(self, other):
        if not isinstance(other, ColumnAssignment):
            return False
        return (
            self.alias == other.alias
            and self.concept == other.concept
            and self.modifiers == other.modifiers
        )

    @property
    def is_complete(self) -> bool:
        return Modifier.PARTIAL not in self.modifiers

    @property
    def is_nullable(self) -> bool:
        return Modifier.NULLABLE in self.modifiers

    def with_namespace(self, namespace: str) -> "ColumnAssignment":
        return ColumnAssignment.model_construct(
            alias=(
                self.alias.with_namespace(namespace)
                if isinstance(self.alias, Function)
                else self.alias
            ),
            concept=self.concept.with_namespace(namespace),
            modifiers=self.modifiers,
        )

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> "ColumnAssignment":
        return ColumnAssignment.model_construct(
            alias=self.alias,
            concept=self.concept.with_merge(source, target, modifiers),
            modifiers=(
                modifiers if self.concept.address == source.address else self.modifiers
            ),
        )


class Address(BaseModel):
    location: str
    is_query: bool = False
    quoted: bool = False


class Query(BaseModel):
    text: str


class DatasourceMetadata(BaseModel):
    freshness_concept: Concept | None
    partition_fields: List[Concept] = Field(default_factory=list)
    line_no: int | None = None


def safe_grain(v) -> Grain:
    if isinstance(v, dict):
        return Grain.model_validate(v)
    elif isinstance(v, Grain):
        return v
    elif not v:
        return Grain(components=set())
    else:
        raise ValueError(f"Invalid input type to safe_grain {type(v)}")


class Datasource(HasUUID, Namespaced, BaseModel):
    name: str
    columns: List[ColumnAssignment]
    address: Union[Address, str]
    grain: Grain = Field(
        default_factory=lambda: Grain(components=set()), validate_default=True
    )
    namespace: Optional[str] = Field(default=DEFAULT_NAMESPACE, validate_default=True)
    metadata: DatasourceMetadata = Field(
        default_factory=lambda: DatasourceMetadata(freshness_concept=None)
    )
    where: Optional[WhereClause] = None
    non_partial_for: Optional[WhereClause] = None

    @property
    def safe_address(self) -> str:
        if isinstance(self.address, Address):
            return self.address.location
        return self.address

    def __eq__(self, other):
        if not isinstance(other, Datasource):
            return False
        return (
            self.name == other.name
            and self.namespace == other.namespace
            and self.grain == other.grain
            and self.address == other.address
            and self.where == other.where
            and self.columns == other.columns
            and self.non_partial_for == other.non_partial_for
        )

    def duplicate(self) -> "Datasource":
        return self.model_copy(deep=True)

    @property
    def hidden_concepts(self) -> List[Concept]:
        return []

    def merge_concept(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ):
        original = [c for c in self.columns if c.concept.address == source.address]
        early_exit_check = [
            c for c in self.columns if c.concept.address == target.address
        ]
        if early_exit_check:
            logger.info(
                f"No concept merge needed on merge of {source} to {target}, have {[x.concept.address for x in self.columns]}"
            )
            return None
        if len(original) != 1:
            raise ValueError(
                f"Expected exactly one column to merge, got {len(original)} for {source.address}, {[x.alias for x in original]}"
            )
        # map to the alias with the modifier, and the original
        self.columns = [
            c.with_merge(source, target, modifiers)
            for c in self.columns
            if c.concept.address != source.address
        ] + original
        self.grain = self.grain.with_merge(source, target, modifiers)
        self.where = (
            self.where.with_merge(source, target, modifiers) if self.where else None
        )

        self.add_column(target, original[0].alias, modifiers)

    @property
    def identifier(self) -> str:
        if not self.namespace or self.namespace == DEFAULT_NAMESPACE:
            return self.name
        return f"{self.namespace}.{self.name}"

    @property
    def safe_identifier(self) -> str:
        return self.identifier.replace(".", "_")

    @property
    def output_lcl(self) -> LooseConceptList:
        return LooseConceptList(concepts=self.output_concepts)

    @property
    def non_partial_concept_addresses(self) -> set[str]:
        return set([c.address for c in self.full_concepts])

    @field_validator("namespace", mode="plain")
    @classmethod
    def namespace_validation(cls, v):
        return v or DEFAULT_NAMESPACE

    @field_validator("address")
    @classmethod
    def address_enforcement(cls, v):
        if isinstance(v, str):
            v = Address(location=v)
        return v

    @field_validator("grain", mode="before")
    @classmethod
    def grain_enforcement(cls, v: Grain, info: ValidationInfo):
        grain: Grain = safe_grain(v)
        return grain

    def add_column(
        self,
        concept: Concept,
        alias: str | RawColumnExpr | Function,
        modifiers: List[Modifier] | None = None,
    ):
        self.columns.append(
            ColumnAssignment(
                alias=alias, concept=concept.reference, modifiers=modifiers or []
            )
        )

    def __add__(self, other):
        if not other == self:
            raise ValueError(
                "Attempted to add two datasources that are not identical, this is not a valid operation"
            )
        return self

    def __repr__(self):
        return f"Datasource<{self.identifier}@<{self.grain}>"

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        return self.identifier.__hash__()

    def with_namespace(self, namespace: str):
        new_namespace = (
            namespace + "." + self.namespace
            if self.namespace and self.namespace != DEFAULT_NAMESPACE
            else namespace
        )
        new = Datasource.model_construct(
            name=self.name,
            namespace=new_namespace,
            grain=self.grain.with_namespace(namespace),
            address=self.address,
            columns=[c.with_namespace(namespace) for c in self.columns],
            where=self.where.with_namespace(namespace) if self.where else None,
            non_partial_for=(
                self.non_partial_for.with_namespace(namespace)
                if self.non_partial_for
                else None
            ),
        )
        return new

    @property
    def concepts(self) -> List[ConceptRef]:
        return [c.concept for c in self.columns]

    @property
    def group_required(self):
        return False

    @property
    def full_concepts(self) -> List[ConceptRef]:
        return [c.concept for c in self.columns if Modifier.PARTIAL not in c.modifiers]

    @property
    def nullable_concepts(self) -> List[ConceptRef]:
        return [c.concept for c in self.columns if Modifier.NULLABLE in c.modifiers]

    @property
    def output_concepts(self) -> List[ConceptRef]:
        return self.concepts

    @property
    def partial_concepts(self) -> List[ConceptRef]:
        return [c.concept for c in self.columns if Modifier.PARTIAL in c.modifiers]


class EnvironmentDatasourceDict(dict):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(self, *args, **kwargs)

    def __getitem__(self, key: str) -> Datasource:
        try:
            return super(EnvironmentDatasourceDict, self).__getitem__(key)
        except KeyError:
            if DEFAULT_NAMESPACE + "." + key in self:
                return self.__getitem__(DEFAULT_NAMESPACE + "." + key)
            if "." in key and key.split(".", 1)[0] == DEFAULT_NAMESPACE:
                return self.__getitem__(key.split(".", 1)[1])
            raise

    def values(self) -> ValuesView[Datasource]:  # type: ignore
        return super().values()

    def items(self) -> ItemsView[str, Datasource]:  # type: ignore
        return super().items()

    def duplicate(self) -> "EnvironmentDatasourceDict":
        new = EnvironmentDatasourceDict()
        new.update({k: v.duplicate() for k, v in self.items()})
        return new



from typing import List

from trilogy.core.enums import SourceType
from trilogy.core.models.build import BuildConcept, BuildFunction
from trilogy.core.models.execute import QueryDatasource, UnnestJoin
from trilogy.core.processing.nodes.base_node import StrategyNode


class UnnestNode(StrategyNode):
    """Unnest nodes represent an expansion of an array or other
    column into rows.
    """

    source_type = SourceType.UNNEST

    def __init__(
        self,
        unnest_concepts: List[BuildConcept],
        input_concepts: List[BuildConcept],
        output_concepts: List[BuildConcept],
        environment,
        whole_grain: bool = False,
        parents: List["StrategyNode"] | None = None,
        depth: int = 0,
    ):
        super().__init__(
            input_concepts=input_concepts,
            output_concepts=output_concepts,
            environment=environment,
            whole_grain=whole_grain,
            parents=parents,
            depth=depth,
        )
        self.unnest_concepts = unnest_concepts

    def _resolve(self) -> QueryDatasource:
        """We need to ensure that any filtered values are removed from the output to avoid inappropriate references"""
        base = super()._resolve()
        lineage = self.unnest_concepts[0].lineage
        assert isinstance(lineage, BuildFunction)
        final = "_".join(set([c.address for c in self.unnest_concepts]))
        unnest = UnnestJoin(
            concepts=self.unnest_concepts,
            parent=lineage,
            alias=f'unnest_{final.replace(".", "_")}',
        )
        base.joins.append(unnest)
        for unnest_concept in self.unnest_concepts:
            base.source_map[unnest_concept.address] = {unnest}
            base.join_derived_concepts = [unnest_concept]
        return base

    def copy(self) -> "UnnestNode":
        return UnnestNode(
            unnest_concepts=self.unnest_concepts,
            input_concepts=list(self.input_concepts),
            output_concepts=list(self.output_concepts),
            environment=self.environment,
            whole_grain=self.whole_grain,
            parents=self.parents,
            depth=self.depth,
        )



from trilogy import Environment, Executor
from trilogy.authoring import DataType, Function
from trilogy.core.enums import FunctionType, Purpose, ValidationScope
from trilogy.core.exceptions import (
    ModelValidationError,
)
from trilogy.core.validation.common import ValidationTest
from trilogy.core.validation.concept import validate_concept
from trilogy.core.validation.datasource import validate_datasource
from trilogy.parsing.common import function_to_concept


def validate_environment(
    env: Environment,
    scope: ValidationScope = ValidationScope.ALL,
    targets: list[str] | None = None,
    exec: Executor | None = None,
    generate_only: bool = False,
) -> list[ValidationTest]:
    # avoid mutating the environment for validation
    generate_only = exec is None or generate_only
    env = env.duplicate()
    grain_check = function_to_concept(
        parent=Function(
            operator=FunctionType.SUM,
            arguments=[1],
            output_datatype=DataType.INTEGER,
            output_purpose=Purpose.METRIC,
        ),
        name="grain_check",
        environment=env,
    )
    env.add_concept(grain_check)
    new_concepts = []
    for concept in env.concepts.values():
        concept_grain_check = function_to_concept(
            parent=Function(
                operator=FunctionType.COUNT_DISTINCT,
                arguments=[concept.reference],
                output_datatype=DataType.INTEGER,
                output_purpose=Purpose.METRIC,
            ),
            name=f"grain_check_{concept.safe_address}",
            environment=env,
        )
        new_concepts.append(concept_grain_check)
    for concept in new_concepts:
        env.add_concept(concept)
    build_env = env.materialize_for_select()
    results: list[ValidationTest] = []
    if scope == ValidationScope.ALL or scope == ValidationScope.DATASOURCES:
        for datasource in build_env.datasources.values():
            if targets and datasource.name not in targets:
                continue
            results += validate_datasource(datasource, env, build_env, exec)
    if scope == ValidationScope.ALL or scope == ValidationScope.CONCEPTS:

        for bconcept in build_env.concepts.values():
            if targets and bconcept.address not in targets:
                continue
            results += validate_concept(bconcept, env, build_env, exec)

    # raise a nicely formatted union of all exceptions
    exceptions: list[ModelValidationError] = [e.result for e in results if e.result]
    if exceptions:
        if not generate_only:
            messages = "\n".join([str(e) for e in exceptions])
            raise ModelValidationError(
                f"Environment validation failed with the following errors:\n{messages}",
                children=exceptions,
            )

    return results



from pytest import raises

from trilogy import Dialects
from trilogy.core.exceptions import UnresolvableQueryException


def test_cannot_find():
    x = Dialects.DUCK_DB.default_executor()

    with raises(UnresolvableQueryException):
        x.generate_sql(
            """
    key x int;
    key y int;

    datasource fun (
    y: y,
        )
    address abc;


    select x;

    """
        )


def test_cannot_find_complex():
    x = Dialects.DUCK_DB.default_executor()
    with raises(UnresolvableQueryException):
        x.generate_sql(
            """
    key x int;
    key y int;

    auto sum <- x+y;

    datasource fun (
    y: y,
        )
    address abc;


    select sum(y) by x as fun;

    """
        )



from collections import defaultdict
from enum import Enum
from typing import List

from trilogy.core.models.build import (
    BuildConcept,
    BuildWhereClause,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.processing.nodes import (
    StrategyNode,
)
from trilogy.core.processing.utility import (
    get_disconnected_components,
)


class ValidationResult(Enum):
    COMPLETE = 1
    DISCONNECTED = 2
    INCOMPLETE = 3
    INCOMPLETE_CONDITION = 4


def validate_concept(
    concept: BuildConcept,
    node: StrategyNode,
    found_addresses: set[str],
    non_partial_addresses: set[str],
    partial_addresses: set[str],
    virtual_addresses: set[str],
    found_map: dict[str, set[BuildConcept]],
    accept_partial: bool,
    seen: set[str],
    environment: BuildEnvironment,
):
    # logger.debug(
    #     f"Validating concept {concept.address} with accept_partial={accept_partial}"
    # )
    found_map[str(node)].add(concept)
    seen.add(concept.address)
    if concept not in node.partial_concepts:
        found_addresses.add(concept.address)
        non_partial_addresses.add(concept.address)
        # remove it from our partial tracking
        if concept.address in partial_addresses:
            partial_addresses.remove(concept.address)
        if concept.address in virtual_addresses:
            virtual_addresses.remove(concept.address)
    if concept in node.partial_concepts:
        if concept.address in non_partial_addresses:
            return None
        partial_addresses.add(concept.address)
        if accept_partial:
            found_addresses.add(concept.address)
            found_map[str(node)].add(concept)
    for v_address in concept.pseudonyms:
        if v_address in seen:
            continue
        if v_address in environment.alias_origin_lookup:
            # logger.debug(
            #     f"Found alias origin for {v_address}: {environment.alias_origin_lookup[v_address]} mapped to {environment.concepts[v_address]}")
            v = environment.alias_origin_lookup[v_address]
        else:
            v = environment.concepts[v_address]

        if v.address in seen:

            continue

        if v.address == concept.address:

            continue
        validate_concept(
            v,
            node,
            found_addresses,
            non_partial_addresses,
            partial_addresses,
            virtual_addresses,
            found_map,
            accept_partial,
            seen=seen,
            environment=environment,
        )


def validate_stack(
    environment: BuildEnvironment,
    stack: List[StrategyNode],
    concepts: List[BuildConcept],
    mandatory_with_filter: List[BuildConcept],
    conditions: BuildWhereClause | None = None,
    accept_partial: bool = False,
) -> tuple[ValidationResult, set[str], set[str], set[str], set[str]]:
    found_map: dict[str, set[BuildConcept]] = defaultdict(set)
    found_addresses: set[str] = set()
    non_partial_addresses: set[str] = set()
    partial_addresses: set[str] = set()
    virtual_addresses: set[str] = set()
    seen: set[str] = set()

    for node in stack:
        resolved = node.resolve()

        for concept in resolved.output_concepts:
            if concept.address in resolved.hidden_concepts:
                continue

            validate_concept(
                concept,
                node,
                found_addresses,
                non_partial_addresses,
                partial_addresses,
                virtual_addresses,
                found_map,
                accept_partial,
                seen,
                environment,
            )
        for concept in node.virtual_output_concepts:
            if concept.address in non_partial_addresses:
                continue
            found_addresses.add(concept.address)
            virtual_addresses.add(concept.address)
    if not conditions:
        conditions_met = True
    else:
        conditions_met = all(
            [node.preexisting_conditions == conditions.conditional for node in stack]
        ) or all([c.address in found_addresses for c in mandatory_with_filter])
    # zip in those we know we found
    if not all([c.address in found_addresses for c in concepts]) or not conditions_met:
        if not all([c.address in found_addresses for c in concepts]):
            return (
                ValidationResult.INCOMPLETE,
                found_addresses,
                {c.address for c in concepts if c.address not in found_addresses},
                partial_addresses,
                virtual_addresses,
            )
        return (
            ValidationResult.INCOMPLETE_CONDITION,
            found_addresses,
            {c.address for c in concepts if c.address not in mandatory_with_filter},
            partial_addresses,
            virtual_addresses,
        )

    graph_count, _ = get_disconnected_components(found_map)
    if graph_count in (0, 1):
        return (
            ValidationResult.COMPLETE,
            found_addresses,
            set(),
            partial_addresses,
            virtual_addresses,
        )
    # if we have too many subgraphs, we need to keep searching
    return (
        ValidationResult.DISCONNECTED,
        found_addresses,
        set(),
        partial_addresses,
        virtual_addresses,
    )



from trilogy.core.enums import Boolean, BooleanOperator


def test_boolean_operator():
    assert BooleanOperator("AND") == BooleanOperator.AND
    assert BooleanOperator("and") == BooleanOperator.AND


def test_boolean():
    assert Boolean("TRUE") == Boolean.TRUE
    assert Boolean("true") == Boolean.TRUE
    assert Boolean(True) == Boolean.TRUE



from trilogy import parse


def test_metadata():
    env, _ = parse(
        """key user_id int metadata(description="the description");
property user_id.display_name string metadata(description="The display name");"""
    )

    assert env.concepts["user_id"].metadata.description == "the description"
    assert env.concepts["display_name"].metadata.description == "The display name"


# def test_import_metadata():
#     env = Environment(working_path=Path(__file__).parent)
#     env, _ = parse(
#         """import test_env as env; # Dragon metrics
#         import test_env as env2;""",
#         environment=env,
#     )

#     assert "Dragon metrics" in env.concepts["env.id"].metadata.description

#     env2 = env.concepts["env2.id"]
#     assert env2.namespace == "env2"
#     assert env.concepts["env2.id"].metadata.description is None



from typing import List

from trilogy.core.enums import (
    SourceType,
)
from trilogy.core.models.build import (
    BuildComparison,
    BuildConcept,
    BuildConditional,
    BuildGrain,
    BuildParenthetical,
)
from trilogy.core.processing.nodes.base_node import StrategyNode


class FilterNode(StrategyNode):
    """Filter nodes represent a restriction operation
    on a concept that creates a new derived concept.

    They should only output a concept and it's filtered
    version, but will have parents that provide all required
    filtering keys as inputs.
    """

    source_type = SourceType.FILTER

    def __init__(
        self,
        input_concepts: List[BuildConcept],
        output_concepts: List[BuildConcept],
        environment,
        whole_grain: bool = False,
        parents: List["StrategyNode"] | None = None,
        depth: int = 0,
        conditions: (
            BuildConditional | BuildComparison | BuildParenthetical | None
        ) = None,
        preexisting_conditions: (
            BuildConditional | BuildComparison | BuildParenthetical | None
        ) = None,
        partial_concepts: List[BuildConcept] | None = None,
        force_group: bool | None = False,
        grain: BuildGrain | None = None,
        existence_concepts: List[BuildConcept] | None = None,
    ):
        super().__init__(
            output_concepts=output_concepts,
            environment=environment,
            whole_grain=whole_grain,
            parents=parents,
            depth=depth,
            input_concepts=input_concepts,
            conditions=conditions,
            preexisting_conditions=preexisting_conditions,
            partial_concepts=partial_concepts,
            force_group=force_group,
            grain=grain,
            existence_concepts=existence_concepts,
        )

    def copy(self) -> "FilterNode":
        return FilterNode(
            input_concepts=list(self.input_concepts),
            output_concepts=list(self.output_concepts),
            environment=self.environment,
            whole_grain=self.whole_grain,
            parents=self.parents,
            depth=self.depth,
            conditions=self.conditions,
            preexisting_conditions=self.preexisting_conditions,
            partial_concepts=list(self.partial_concepts),
            force_group=self.force_group,
            grain=self.grain,
            existence_concepts=list(self.existence_concepts),
        )



from dataclasses import dataclass
from typing import List, Optional

from trilogy.constants import logger
from trilogy.core.enums import Derivation, Granularity
from trilogy.core.env_processor import generate_graph
from trilogy.core.exceptions import UnresolvableQueryException
from trilogy.core.graph_models import ReferenceGraph
from trilogy.core.models.author import (
    UndefinedConcept,
)
from trilogy.core.models.build import (
    BuildConcept,
    BuildWhereClause,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.processing.discovery_node_factory import generate_node
from trilogy.core.processing.discovery_utility import (
    LOGGER_PREFIX,
    depth_to_prefix,
    get_priority_concept,
    group_if_required_v2,
)
from trilogy.core.processing.discovery_validation import (
    ValidationResult,
    validate_stack,
)
from trilogy.core.processing.nodes import (
    History,
    MergeNode,
    StrategyNode,
)
from trilogy.utility import unique

SKIPPED_DERIVATIONS = [
    Derivation.AGGREGATE,
    Derivation.FILTER,
    Derivation.WINDOW,
    Derivation.UNNEST,
    Derivation.RECURSIVE,
    Derivation.ROWSET,
    Derivation.BASIC,
    Derivation.GROUP_TO,
    Derivation.MULTISELECT,
    Derivation.UNION,
]

ROOT_DERIVATIONS = [Derivation.ROOT, Derivation.CONSTANT]


def generate_candidates_restrictive(
    priority_concept: BuildConcept,
    candidates: list[BuildConcept],
    exhausted: set[str],
    depth: int,
    conditions: BuildWhereClause | None = None,
) -> tuple[list[BuildConcept], BuildWhereClause | None]:
    local_candidates = [
        x
        for x in list(candidates)
        if x.address not in exhausted
        and x.granularity != Granularity.SINGLE_ROW
        and x.address not in priority_concept.pseudonyms
        and priority_concept.address not in x.pseudonyms
    ]

    # if it's single row, joins are irrelevant. Fetch without keys.
    if priority_concept.granularity == Granularity.SINGLE_ROW:
        logger.info("Have single row concept, including only other single row optional")
        optional = (
            [
                x
                for x in candidates
                if x.granularity == Granularity.SINGLE_ROW
                and x.address not in priority_concept.pseudonyms
                and priority_concept.address not in x.pseudonyms
            ]
            if priority_concept.derivation == Derivation.AGGREGATE
            else []
        )
        return optional, conditions

    if conditions and priority_concept.derivation in ROOT_DERIVATIONS:
        logger.info(
            f"{depth_to_prefix(depth)}{LOGGER_PREFIX} Injecting additional conditional row arguments as all remaining concepts are roots or constant"
        )
        # otherwise, we can ignore the conditions now that we've injected inputs
        return (
            unique(list(conditions.row_arguments) + local_candidates, "address"),
            None,
        )

    return local_candidates, conditions


def append_existence_check(
    node: StrategyNode,
    environment: BuildEnvironment,
    graph: ReferenceGraph,
    where: BuildWhereClause,
    history: History,
):
    # we if we have a where clause doing an existence check
    # treat that as separate subquery
    if where.existence_arguments:
        for subselect in where.existence_arguments:
            if not subselect:
                continue
            if all([x.address in node.input_concepts for x in subselect]):
                logger.info(
                    f"{LOGGER_PREFIX} existance clause inputs already found {[str(c) for c in subselect]}"
                )
                continue
            logger.info(
                f"{LOGGER_PREFIX} fetching existence clause inputs {[str(c) for c in subselect]}"
            )
            parent = source_query_concepts(
                [*subselect],
                history=history,
                environment=environment,
                g=graph,
            )
            assert parent, "Could not resolve existence clause"
            node.add_parents([parent])
            logger.info(f"{LOGGER_PREFIX} found {[str(c) for c in subselect]}")
            node.add_existence_concepts([*subselect])


def search_concepts(
    mandatory_list: List[BuildConcept],
    history: History,
    environment: BuildEnvironment,
    depth: int,
    g: ReferenceGraph,
    accept_partial: bool = False,
    conditions: BuildWhereClause | None = None,
) -> StrategyNode | None:
    hist = history.get_history(
        search=mandatory_list, accept_partial=accept_partial, conditions=conditions
    )
    if hist is not False:
        logger.info(
            f"{depth_to_prefix(depth)}{LOGGER_PREFIX} Returning search node from history ({'exists' if hist is not None else 'does not exist'}) for {[c.address for c in mandatory_list]} with accept_partial {accept_partial}"
        )
        assert not isinstance(hist, bool)
        return hist

    result = _search_concepts(
        mandatory_list,
        environment,
        depth=depth,
        g=g,
        accept_partial=accept_partial,
        history=history,
        conditions=conditions,
    )
    # a node may be mutated after be cached; always store a copy
    history.search_to_history(
        mandatory_list,
        accept_partial,
        result.copy() if result else None,
        conditions=conditions,
    )
    return result


@dataclass
class LoopContext:
    mandatory_list: List[BuildConcept]
    environment: BuildEnvironment
    depth: int
    g: ReferenceGraph
    history: History
    attempted: set[str]
    found: set[str]
    skip: set[str]
    all_mandatory: set[str]
    original_mandatory: List[BuildConcept]
    completion_mandatory: List[BuildConcept]
    stack: List[StrategyNode]
    complete: ValidationResult = ValidationResult.INCOMPLETE
    accept_partial: bool = False
    must_evaluate_condition_on_this_level_not_push_down: bool = False
    conditions: BuildWhereClause | None = None

    @property
    def incomplete(self) -> bool:
        return self.attempted != self.all_mandatory


def initialize_loop_context(
    mandatory_list: List[BuildConcept],
    environment: BuildEnvironment,
    depth: int,
    g: ReferenceGraph,
    history: History,
    accept_partial: bool = False,
    conditions: BuildWhereClause | None = None,
):
    # these are the concepts we need in the output projection
    mandatory_list = unique(mandatory_list, "address")
    # cache our values before an filter injection
    original_mandatory = [*mandatory_list]
    for x in mandatory_list:
        if isinstance(x, UndefinedConcept):
            raise SyntaxError(f"Undefined concept {x.address}")
    all_mandatory = set(c.address for c in mandatory_list)

    must_evaluate_condition_on_this_level_not_push_down = False

    # if we have a filter, we may need to get more values to support that.
    if conditions:
        completion_mandatory = unique(
            mandatory_list + list(conditions.row_arguments), "address"
        )
        # if anything we need to get is in the filter set and it's a computed value
        # we need to get _everything_ in this loop
        required_filters = [
            x
            for x in mandatory_list
            if x.derivation not in ROOT_DERIVATIONS
            and not (
                x.derivation == Derivation.AGGREGATE
                and x.granularity == Granularity.SINGLE_ROW
            )
            and x.address in conditions.row_arguments
        ]
        if any(required_filters):
            logger.info(
                f"{depth_to_prefix(depth)}{LOGGER_PREFIX} derived condition row inputs {[x.address for x in required_filters]} present in mandatory list, forcing condition evaluation at this level. "
            )
            mandatory_list = completion_mandatory
            all_mandatory = set(c.address for c in completion_mandatory)
            must_evaluate_condition_on_this_level_not_push_down = True
        else:
            logger.info(
                f"{depth_to_prefix(depth)}{LOGGER_PREFIX} Do not need to evaluate conditions yet."
            )
    else:

        completion_mandatory = mandatory_list
    logger.info(
        f"{depth_to_prefix(depth)}{LOGGER_PREFIX} Initialized loop context with mandatory list {[c.address for c in mandatory_list]} and completion mandatory {[c.address for c in completion_mandatory]}"
    )
    return LoopContext(
        mandatory_list=mandatory_list,
        environment=environment,
        depth=depth,
        g=g,
        history=history,
        attempted=set(),
        found=set(),
        skip=set(),
        all_mandatory=all_mandatory,
        original_mandatory=original_mandatory,
        completion_mandatory=completion_mandatory,
        stack=[],
        complete=ValidationResult.INCOMPLETE,
        accept_partial=accept_partial,
        must_evaluate_condition_on_this_level_not_push_down=must_evaluate_condition_on_this_level_not_push_down,
        conditions=conditions,
    )


def evaluate_loop_conditions(
    context: LoopContext, priority_concept: BuildConcept
) -> BuildWhereClause | None:
    # filter evaluation
    # always pass the filter up when we aren't looking at all filter inputs
    # or there are any non-filter complex types
    if context.conditions:
        should_evaluate_filter_on_this_level_not_push_down = all(
            [
                x.address in context.mandatory_list
                for x in context.conditions.row_arguments
            ]
        ) and not any(
            [
                x.derivation not in ROOT_DERIVATIONS + [Derivation.BASIC]
                for x in context.mandatory_list
                if x.address not in context.conditions.row_arguments
            ]
        )
    else:
        should_evaluate_filter_on_this_level_not_push_down = True
    local_conditions = (
        context.conditions
        if context.conditions
        and not context.must_evaluate_condition_on_this_level_not_push_down
        and not should_evaluate_filter_on_this_level_not_push_down
        else None
    )
    # but if it's not basic, and it's not condition;
    # we do need to push it down (and have another layer of filter evaluation)
    # to ensure filtering happens before something like a SUM
    if (
        context.conditions
        and priority_concept.derivation not in ROOT_DERIVATIONS + [Derivation.BASIC]
        and priority_concept.address not in context.conditions.row_arguments
    ):
        logger.info(
            f"{depth_to_prefix(context.depth)}{LOGGER_PREFIX} Force including conditions in {priority_concept.address} to push filtering above complex condition that is not condition member or parent"
        )
        local_conditions = context.conditions
    return local_conditions


def check_for_early_exit(
    complete, partial, missing, context: LoopContext, priority_concept: BuildConcept
) -> bool:
    if complete == ValidationResult.INCOMPLETE_CONDITION:
        cond_dict = {str(node): node.preexisting_conditions for node in context.stack}
        for node in context.stack:
            logger.info(
                f"{depth_to_prefix(context.depth)}{LOGGER_PREFIX} Node {node} has conditions {node.preexisting_conditions} and {node.conditions}"
            )
        raise SyntaxError(f"Have {cond_dict} and need {str(context.conditions)}")
    # early exit if we have a complete stack with one node
    # we can only early exit if we have a complete stack
    # and we are not looking for more non-partial sources
    if complete == ValidationResult.COMPLETE and (
        not context.accept_partial or (context.accept_partial and not partial)
    ):
        logger.info(
            f"{depth_to_prefix(context.depth)}{LOGGER_PREFIX} breaking loop, complete"
        )
        return True
    elif complete == ValidationResult.COMPLETE and context.accept_partial and partial:
        if len(context.attempted) == len(context.mandatory_list):
            logger.info(
                f"{depth_to_prefix(context.depth)}{LOGGER_PREFIX} Breaking as we have attempted all nodes"
            )
            return True
        logger.info(
            f"{depth_to_prefix(context.depth)}{LOGGER_PREFIX} Found complete stack with partials {partial}, continuing search, attempted {context.attempted} all {len(context.mandatory_list)}"
        )
    else:
        logger.info(
            f"{depth_to_prefix(context.depth)}{LOGGER_PREFIX} Not complete (missing {missing}), continuing search"
        )
    # if we have attempted on root node, we've tried them all.
    if priority_concept.derivation == Derivation.ROOT:
        logger.info(
            f"{depth_to_prefix(context.depth)}{LOGGER_PREFIX} Breaking as attempted root with no results"
        )
        return True
    return False


def generate_loop_completion(context: LoopContext, virtual: set[str]) -> StrategyNode:
    condition_required = True
    non_virtual = [c for c in context.completion_mandatory if c.address not in virtual]
    non_virtual_output = [
        c for c in context.original_mandatory if c.address not in virtual
    ]
    non_virtual_different = len(context.completion_mandatory) != len(
        context.original_mandatory
    )
    non_virtual_difference_values = set(
        [x.address for x in context.completion_mandatory]
    ).difference(set([x.address for x in context.original_mandatory]))
    if not context.conditions:
        condition_required = False
        non_virtual = [c for c in context.mandatory_list if c.address not in virtual]

    elif all(
        [
            x.preexisting_conditions == context.conditions.conditional
            for x in context.stack
        ]
    ):
        condition_required = False
        non_virtual = [c for c in context.mandatory_list if c.address not in virtual]

    if context.conditions and not condition_required:
        parent_map = {
            str(x): x.preexisting_conditions == context.conditions.conditional
            for x in context.stack
        }
        logger.info(
            f"Condition {context.conditions} not required, parents included filtering! {parent_map}"
        )
    if len(context.stack) == 1:
        output: StrategyNode = context.stack[0]
        if non_virtual_different:
            logger.info(
                f"{depth_to_prefix(context.depth)}{LOGGER_PREFIX} Found different non-virtual output concepts ({non_virtual_difference_values}), removing condition injected values by setting outputs to {[x.address for x in output.output_concepts if x.address in non_virtual_output]}"
            )
            # output.set_output_concepts(
            #     [
            #         x
            #         for x in output.output_concepts
            #         if x.address not in non_virtual_difference_values
            #         or any(c in non_virtual_output for c in x.pseudonyms)
            #     ],
            #     rebuild=True,
            #     change_visibility=False
            # )
            # output.set_output_concepts(context.original_mandatory)

            # if isinstance(output, MergeNode):
            #     output.force_group = True
            #     output.rebuild_cache()

        logger.info(
            f"{depth_to_prefix(context.depth)}{LOGGER_PREFIX} Source stack has single node, returning that {type(output)}"
        )
    else:
        logger.info(
            f"{depth_to_prefix(context.depth)}{LOGGER_PREFIX} wrapping multiple parent nodes {[type(x) for x in context.stack]} in merge node"
        )
        output = MergeNode(
            input_concepts=non_virtual,
            output_concepts=non_virtual,
            environment=context.environment,
            parents=context.stack,
            depth=context.depth,
        )

    # ensure we can resolve our final merge
    output.resolve()
    if condition_required and context.conditions:
        output.add_condition(context.conditions.conditional)
        if context.conditions.existence_arguments:
            append_existence_check(
                output,
                context.environment,
                context.g,
                where=context.conditions,
                history=context.history,
            )
    elif context.conditions:
        output.preexisting_conditions = context.conditions.conditional
    logger.info(
        f"{depth_to_prefix(context.depth)}{LOGGER_PREFIX} Graph is connected, returning {type(output)} node output {[x.address for x in output.usable_outputs]} partial {[c.address for c in output.partial_concepts or []]} with {context.conditions}"
    )
    from trilogy.core.processing.discovery_utility import group_if_required_v2

    if condition_required and context.conditions and non_virtual_different:
        logger.info(
            f"{depth_to_prefix(context.depth)}{LOGGER_PREFIX} Conditions {context.conditions} were injected, checking if we need a group to restore grain"
        )
        return group_if_required_v2(
            output, context.original_mandatory, context.environment
        )

    return group_if_required_v2(output, context.original_mandatory, context.environment)


def _search_concepts(
    mandatory_list: List[BuildConcept],
    environment: BuildEnvironment,
    depth: int,
    g: ReferenceGraph,
    history: History,
    accept_partial: bool = False,
    conditions: BuildWhereClause | None = None,
) -> StrategyNode | None:
    # check for direct materialization first
    candidate = history.gen_select_node(
        mandatory_list,
        environment,
        g,
        depth + 1,
        fail_if_not_found=False,
        accept_partial=accept_partial,
        conditions=conditions,
    )

    if candidate:
        return candidate
    context = initialize_loop_context(
        mandatory_list=mandatory_list,
        environment=environment,
        depth=depth,
        g=g,
        history=history,
        accept_partial=accept_partial,
        conditions=conditions,
    )

    while context.incomplete:

        priority_concept = get_priority_concept(
            context.mandatory_list,
            context.attempted,
            found_concepts=context.found,
            depth=depth,
        )

        local_conditions = evaluate_loop_conditions(context, priority_concept)

        candidates = [
            c for c in context.mandatory_list if c.address != priority_concept.address
        ]
        # the local conditions list may be overriden if we end up injecting conditions
        candidate_list, local_conditions = generate_candidates_restrictive(
            priority_concept,
            candidates,
            context.skip,
            depth=depth,
            conditions=local_conditions,
        )

        logger.info(
            f"{depth_to_prefix(depth)}{LOGGER_PREFIX} priority concept is {str(priority_concept)} derivation {priority_concept.derivation} granularity {priority_concept.granularity} with conditions {local_conditions}"
        )

        logger.info(
            f"{depth_to_prefix(depth)}{LOGGER_PREFIX} Beginning sourcing loop for {priority_concept.address}, accept_partial {accept_partial}, optional {[v.address for v in candidate_list]}, exhausted {[c for c in context.skip]}"
        )
        node = generate_node(
            priority_concept,
            candidate_list,
            environment,
            g,
            depth,
            source_concepts=search_concepts,
            accept_partial=accept_partial,
            history=history,
            conditions=local_conditions,
        )
        if node:
            context.stack.append(node)
            node.resolve()
            # these concepts should not be attempted to be sourced again
            # as fetching them requires operating on a subset of concepts
            if priority_concept.derivation in SKIPPED_DERIVATIONS:
                context.skip.add(priority_concept.address)
        context.attempted.add(priority_concept.address)
        complete, found_c, missing_c, partial, virtual = validate_stack(
            environment,
            context.stack,
            context.mandatory_list,
            context.completion_mandatory,
            conditions=context.conditions,
            accept_partial=accept_partial,
        )
        # assign
        context.found = found_c
        early_exit = check_for_early_exit(
            complete, partial, missing_c, context, priority_concept
        )
        if early_exit:
            break

    logger.info(
        f"{depth_to_prefix(depth)}{LOGGER_PREFIX} finished sourcing loop (complete: {complete}), have {context.found} from {[n for n in context.stack]} (missing {context.all_mandatory - context.found}), attempted {context.attempted}, virtual {virtual}"
    )
    if complete == ValidationResult.COMPLETE:
        return generate_loop_completion(context, virtual)

    # if we can't find it after expanding to a merge, then
    # accept partials in join paths
    if not accept_partial:
        logger.info(
            f"{depth_to_prefix(depth)}{LOGGER_PREFIX} Stack is not connected graph, flag for accepting partial addresses is {accept_partial}, changing flag"
        )
        partial_search = search_concepts(
            # use the original mandatory list
            mandatory_list=mandatory_list,
            environment=environment,
            depth=depth,
            g=g,
            accept_partial=True,
            history=history,
            conditions=conditions,
        )
        if partial_search:
            logger.info(
                f"{depth_to_prefix(depth)}{LOGGER_PREFIX} Found {[c.address for c in mandatory_list]} by accepting partials"
            )
            return partial_search
    logger.error(
        f"{depth_to_prefix(depth)}{LOGGER_PREFIX} Could not resolve concepts {[c.address for c in mandatory_list]}, network outcome was {complete}, missing {context.all_mandatory - context.found},"
    )

    return None


def source_query_concepts(
    output_concepts: List[BuildConcept],
    history: History,
    environment: BuildEnvironment,
    g: Optional[ReferenceGraph] = None,
    conditions: Optional[BuildWhereClause] = None,
):
    if not output_concepts:
        raise ValueError(f"No output concepts provided {output_concepts}")
    if not g:
        g = generate_graph(environment)

    root = search_concepts(
        mandatory_list=output_concepts,
        environment=environment,
        g=g,
        depth=0,
        history=history,
        conditions=conditions,
    )
    if not root:
        error_strings = [
            f"{c.address}<{c.purpose}>{c.derivation}>" for c in output_concepts
        ]
        raise UnresolvableQueryException(
            f"Could not resolve connections for query with output {error_strings} from current model."
        )
    final = [x for x in root.output_concepts if x.address not in root.hidden_concepts]
    logger.info(
        f"{depth_to_prefix(0)}{LOGGER_PREFIX} final concepts are {[x.address for x in final]}"
    )
    return group_if_required_v2(root, output_concepts, environment)



from trilogy.constants import DEFAULT_NAMESPACE
from trilogy.core.enums import (
    BooleanOperator,
    ComparisonOperator,
    FunctionClass,
    FunctionType,
    InfiniteFunctionArgs,
    Ordering,
    Purpose,
)
from trilogy.core.functions import FunctionFactory
from trilogy.core.models.author import (
    AggregateWrapper,
    CaseElse,
    CaseWhen,
    Comparison,
    Concept,
    ConceptRef,
    Conditional,
    FilterItem,
    Function,
    FunctionCallWrapper,
    HavingClause,
    MagicConstants,
    Metadata,
    MultiSelectLineage,
    OrderBy,
    OrderItem,
    Parenthetical,
    RowsetItem,
    SubselectComparison,
    WhereClause,
    WindowItem,
    WindowItemOrder,
    WindowItemOver,
    WindowOrder,
    WindowType,
)
from trilogy.core.models.core import (
    ArrayType,
    DataType,
    ListWrapper,
    MapType,
    NumericType,
    StructType,
    TraitDataType,
)
from trilogy.core.models.datasource import Address, Datasource, DatasourceMetadata
from trilogy.core.models.environment import Environment
from trilogy.core.statements.author import (
    ConceptDeclarationStatement,
    ConceptTransform,
    CopyStatement,
    Grain,
    HasUUID,
    ImportStatement,
    MultiSelectStatement,
    PersistStatement,
    RawSQLStatement,
    RowsetDerivationStatement,
    SelectItem,
    SelectStatement,
    ShowCategory,
    ShowStatement,
    ValidateStatement,
)
from trilogy.parsing.common import arbitrary_to_concept, arg_to_datatype

__all__ = [
    # trilogy.constants
    "DEFAULT_NAMESPACE",
    # trilogy.core.enums
    "BooleanOperator",
    "ComparisonOperator",
    "FunctionClass",
    "FunctionType",
    "InfiniteFunctionArgs",
    "Ordering",
    "Purpose",
    # trilogy.core.functions
    "FunctionFactory",
    # trilogy.core.models.author
    "AggregateWrapper",
    "CaseElse",
    "CaseWhen",
    "Comparison",
    "Concept",
    "ConceptRef",
    "Conditional",
    "FilterItem",
    "Function",
    "FunctionCallWrapper",
    "HavingClause",
    "MagicConstants",
    "Metadata",
    "MultiSelectLineage",
    "OrderBy",
    "OrderItem",
    "Parenthetical",
    "RowsetItem",
    "SubselectComparison",
    "WhereClause",
    "WindowItem",
    "WindowItemOrder",
    "WindowItemOver",
    "WindowOrder",
    "WindowType",
    # trilogy.core.models.core
    "ArrayType",
    "DataType",
    "ListWrapper",
    "MapType",
    "NumericType",
    "StructType",
    "TraitDataType",
    # trilogy.core.models.datasource
    "Address",
    "Datasource",
    "DatasourceMetadata",
    # trilogy.core.models.environment
    "Environment",
    # trilogy.core.statements.author
    "ConceptDeclarationStatement",
    "ConceptTransform",
    "CopyStatement",
    "Grain",
    "HasUUID",
    "ImportStatement",
    "MultiSelectStatement",
    "PersistStatement",
    "RawSQLStatement",
    "RowsetDerivationStatement",
    "SelectItem",
    "SelectStatement",
    "ShowCategory",
    "ShowStatement",
    "ValidateStatement",
    # trilogy.parsing.common
    "arbitrary_to_concept",
    "arg_to_datatype",
]



from trilogy.constants import DEFAULT_NAMESPACE
from trilogy.core.enums import ConceptSource, DatePart, FunctionType, Purpose
from trilogy.core.functions import AttrAccess
from trilogy.core.models.author import Concept, Function, Grain, Metadata, TraitDataType
from trilogy.core.models.core import DataType, StructType, arg_to_datatype
from trilogy.core.models.environment import Environment
from trilogy.parsing.common import Meta


def generate_date_concepts(concept: Concept, environment: Environment):
    if concept.metadata and concept.metadata.line_number:
        base_line_number = concept.metadata.line_number
    else:
        base_line_number = None
    arg_tuples: list[tuple[FunctionType, TraitDataType]] = [
        (FunctionType.MONTH, TraitDataType(type=DataType.INTEGER, traits=["month"])),
        (FunctionType.YEAR, TraitDataType(type=DataType.INTEGER, traits=["year"])),
        (
            FunctionType.QUARTER,
            TraitDataType(type=DataType.INTEGER, traits=["quarter"]),
        ),
        (FunctionType.DAY, TraitDataType(type=DataType.INTEGER, traits=["day"])),
        (
            FunctionType.DAY_OF_WEEK,
            TraitDataType(type=DataType.INTEGER, traits=["day_of_week"]),
        ),
    ]
    for ftype, dtype in arg_tuples:
        fname = ftype.name.lower()
        address = concept.address + f".{fname}"
        if address in environment.concepts:
            continue
        default_type = (
            Purpose.CONSTANT
            if concept.purpose == Purpose.CONSTANT
            else Purpose.PROPERTY
        )
        function = Function.model_construct(
            operator=ftype,
            arguments=[concept.reference],
            output_datatype=dtype,
            output_purpose=default_type,
        )
        new_concept = Concept.model_construct(
            name=f"{concept.name}.{fname}",
            datatype=function.output_datatype,
            purpose=default_type,
            lineage=function,
            grain=concept.grain,
            namespace=concept.namespace,
            keys=set(
                [concept.address],
            ),
            metadata=Metadata(
                line_number=base_line_number,
                concept_source=ConceptSource.AUTO_DERIVED,
            ),
        )
        environment.add_concept(new_concept, add_derived=False)
    for grain in [DatePart.MONTH, DatePart.YEAR]:
        address = concept.address + f".{grain.value}_start"
        if address in environment.concepts:
            continue
        function = Function.model_construct(
            operator=FunctionType.DATE_TRUNCATE,
            arguments=[concept.reference, grain],
            output_datatype=DataType.DATE,
            output_purpose=default_type,
            arg_count=2,
        )
        new_concept = Concept.model_construct(
            name=f"{concept.name}.{grain.value}_start",
            datatype=DataType.DATE,
            purpose=Purpose.PROPERTY,
            lineage=function,
            grain=concept.grain.model_copy(),
            namespace=concept.namespace,
            keys=set(
                [concept.address],
            ),
            metadata=Metadata(
                # description=f"Auto-derived from {base_description}. The date truncated to the {grain.value}.",
                line_number=base_line_number,
                concept_source=ConceptSource.AUTO_DERIVED,
            ),
        )

        environment.add_concept(new_concept, add_derived=False)


def generate_datetime_concepts(concept: Concept, environment: Environment):
    if concept.metadata and concept.metadata.line_number:
        base_line_number = concept.metadata.line_number
    else:
        base_line_number = None
    setup_tuples: list[tuple[FunctionType, DataType | TraitDataType]] = [
        (FunctionType.DATE, DataType.DATE),
        (FunctionType.HOUR, TraitDataType(type=DataType.INTEGER, traits=["hour"])),
        (FunctionType.MINUTE, TraitDataType(type=DataType.INTEGER, traits=["minute"])),
        (FunctionType.SECOND, TraitDataType(type=DataType.INTEGER, traits=["second"])),
    ]
    for ftype, datatype in setup_tuples:
        fname = ftype.name.lower()
        address = concept.address + f".{fname}"
        if address in environment.concepts:
            continue
        default_type = (
            Purpose.CONSTANT
            if concept.purpose == Purpose.CONSTANT
            else Purpose.PROPERTY
        )
        const_function = Function.model_construct(
            operator=ftype,
            arguments=[concept.reference],
            output_datatype=datatype,
            output_purpose=default_type,
        )
        new_concept = Concept.model_construct(
            name=f"{concept.name}.{fname}",
            datatype=datatype,
            purpose=default_type,
            lineage=const_function,
            grain=concept.grain.model_copy(),
            namespace=concept.namespace,
            keys=set(
                [concept.address],
            ),
            metadata=Metadata(
                line_number=base_line_number,
                concept_source=ConceptSource.AUTO_DERIVED,
            ),
        )
        if new_concept.name in environment.concepts:
            continue
        environment.add_concept(new_concept, add_derived=False)


def generate_key_concepts(concept: Concept, environment: Environment):
    if concept.metadata and concept.metadata.line_number:
        base_line_number = concept.metadata.line_number
    else:
        base_line_number = None
    for ftype in [FunctionType.COUNT]:
        address = concept.address + f".{ftype.name.lower()}"
        if address in environment.concepts:
            continue
        fname = ftype.name.lower()
        default_type = Purpose.METRIC
        const_function: Function = Function.model_construct(
            operator=ftype,
            output_datatype=DataType.INTEGER,
            output_purpose=default_type,
            arguments=[concept.reference],
        )
        new_concept = Concept.model_construct(
            name=f"{concept.name}.{fname}",
            datatype=DataType.INTEGER,
            purpose=default_type,
            lineage=const_function,
            grain=Grain(),
            namespace=concept.namespace,
            keys=set(),
            metadata=Metadata(
                # description=f"Auto-derived integer. The {ftype.value} of {concept.address}, {base_description}",
                line_number=base_line_number,
                concept_source=ConceptSource.AUTO_DERIVED,
            ),
        )
        environment.add_concept(new_concept, add_derived=False)


def remove_date_concepts(concept: Concept, environment: Environment):
    """Remove auto-generated date-related concepts for the given concept"""
    date_suffixes = ["month", "year", "quarter", "day", "day_of_week"]
    grain_suffixes = ["month_start", "year_start"]

    for suffix in date_suffixes + grain_suffixes:
        address = concept.address + f".{suffix}"
        if address in environment.concepts:
            derived_concept = environment.concepts[address]
            # Only remove if it was auto-derived from this concept
            if (
                derived_concept.metadata
                and derived_concept.metadata.concept_source
                == ConceptSource.AUTO_DERIVED
                and derived_concept.keys
                and concept.address in derived_concept.keys
            ):
                environment.remove_concept(address)


def remove_datetime_concepts(concept: Concept, environment: Environment):
    """Remove auto-generated datetime-related concepts for the given concept"""
    datetime_suffixes = ["date", "hour", "minute", "second"]

    for suffix in datetime_suffixes:
        address = concept.address + f".{suffix}"
        if address in environment.concepts:
            derived_concept = environment.concepts[address]
            # Only remove if it was auto-derived from this concept
            if (
                derived_concept.metadata
                and derived_concept.metadata.concept_source
                == ConceptSource.AUTO_DERIVED
                and derived_concept.keys
                and concept.address in derived_concept.keys
            ):
                environment.remove_concept(address)


def remove_key_concepts(concept: Concept, environment: Environment):
    """Remove auto-generated key-related concepts for the given concept"""
    key_suffixes = ["count"]

    for suffix in key_suffixes:
        address = concept.address + f".{suffix}"
        if address in environment.concepts:
            derived_concept = environment.concepts[address]
            if (
                derived_concept.metadata
                and derived_concept.metadata.concept_source
                == ConceptSource.AUTO_DERIVED
            ):
                environment.remove_concept(address)


def remove_struct_concepts(concept: Concept, environment: Environment):
    """Remove auto-generated struct field concepts for the given concept"""
    if not isinstance(concept.datatype, StructType):
        return

    target_namespace = (
        environment.namespace + "." + concept.name
        if environment.namespace and environment.namespace != DEFAULT_NAMESPACE
        else concept.name
    )

    # Get all concepts in the target namespace that were auto-derived
    concepts_to_remove = []
    for address, derived_concept in environment.concepts.items():
        if (
            derived_concept.namespace == target_namespace
            and derived_concept.metadata
            and derived_concept.metadata.concept_source == ConceptSource.AUTO_DERIVED
            and isinstance(derived_concept.lineage, Function)
            and derived_concept.lineage.operator == FunctionType.ATTR_ACCESS
            and len(derived_concept.lineage.arguments) >= 1
            and derived_concept.lineage.arguments[0] == concept.reference
        ):
            concepts_to_remove.append(address)

    for address in concepts_to_remove:
        environment.remove_concept(address)


def remove_related_concepts(concept: Concept, environment: Environment):
    """Remove all auto-generated concepts that were derived from the given concept"""

    # Remove key-related concepts
    if concept.purpose == Purpose.KEY:
        remove_key_concepts(concept, environment)

    # Remove datatype-specific concepts
    if concept.datatype == DataType.DATE:
        remove_date_concepts(concept, environment)
    elif concept.datatype == DataType.DATETIME:
        remove_date_concepts(concept, environment)
        remove_datetime_concepts(concept, environment)
    elif concept.datatype == DataType.TIMESTAMP:
        remove_date_concepts(concept, environment)
        remove_datetime_concepts(concept, environment)

    # Remove struct field concepts
    if isinstance(concept.datatype, StructType):
        remove_struct_concepts(concept, environment)


def generate_related_concepts(
    concept: Concept,
    environment: Environment,
    meta: Meta | None = None,
    add_derived: bool = False,
):
    """Auto populate common derived concepts on types"""
    if concept.purpose == Purpose.KEY and add_derived:
        generate_key_concepts(concept, environment)

    # datatype types
    if concept.datatype == DataType.DATE and add_derived:
        generate_date_concepts(concept, environment)
    elif concept.datatype == DataType.DATETIME and add_derived:

        generate_date_concepts(concept, environment)
        generate_datetime_concepts(concept, environment)
    elif concept.datatype == DataType.TIMESTAMP and add_derived:
        generate_date_concepts(concept, environment)
        generate_datetime_concepts(concept, environment)

    if isinstance(concept.datatype, StructType):
        for key, value in concept.datatype.fields_map.items():
            auto = Concept.model_construct(
                name=key,
                datatype=arg_to_datatype(value),
                purpose=Purpose.PROPERTY,
                namespace=(
                    environment.namespace + "." + concept.name
                    if environment.namespace
                    and environment.namespace != DEFAULT_NAMESPACE
                    else concept.name
                ),
                lineage=AttrAccess([concept.reference, key], environment=environment),
                grain=concept.grain,
                metadata=Metadata(
                    concept_source=ConceptSource.AUTO_DERIVED,
                ),
                keys=concept.keys,
            )
            environment.add_concept(auto, meta=meta)
            if isinstance(value, Concept):
                environment.merge_concept(auto, value, modifiers=[])



from dataclasses import dataclass
from typing import List, Optional, Protocol, Union

from trilogy.constants import logger
from trilogy.core.enums import Derivation, Granularity
from trilogy.core.graph_models import ReferenceGraph
from trilogy.core.models.build import (
    BuildConcept,
    BuildWhereClause,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.processing.discovery_utility import LOGGER_PREFIX, depth_to_prefix
from trilogy.core.processing.node_generators import (
    gen_basic_node,
    gen_constant_node,
    gen_filter_node,
    gen_group_node,
    gen_group_to_node,
    gen_merge_node,
    gen_multiselect_node,
    gen_recursive_node,
    gen_rowset_node,
    gen_synonym_node,
    gen_union_node,
    gen_unnest_node,
    gen_window_node,
)
from trilogy.core.processing.nodes import (
    History,
    StrategyNode,
)


class SearchConceptsType(Protocol):
    def __call__(
        self,
        mandatory_list: List[BuildConcept],
        history: History,
        environment: BuildEnvironment,
        depth: int,
        g: ReferenceGraph,
        accept_partial: bool = False,
        conditions: Optional[BuildWhereClause] = None,
    ) -> Union[StrategyNode, None]: ...


@dataclass
class NodeGenerationContext:
    """Encapsulates common parameters for node generation."""

    concept: BuildConcept
    local_optional: List[BuildConcept]
    environment: BuildEnvironment
    g: ReferenceGraph
    depth: int
    source_concepts: SearchConceptsType
    history: History
    accept_partial: bool = False
    conditions: Optional[BuildWhereClause] = None

    @property
    def next_depth(self) -> int:
        return self.depth + 1

    def log_generation(self, node_type: str, extra_info: str = "") -> None:
        """Centralized logging for node generation."""
        optional_addresses = [x.address for x in self.local_optional]
        base_msg = f"for {self.concept.address}, generating {node_type} node with optional {optional_addresses}"

        if extra_info:
            base_msg += f" and {extra_info}"

        logger.info(f"{depth_to_prefix(self.depth)}{LOGGER_PREFIX} {base_msg}")


def restrict_node_outputs_targets(
    node: StrategyNode, targets: list[BuildConcept], depth: int
) -> list[BuildConcept]:
    """Restricts node outputs to target concepts and returns extra concepts."""
    ex_resolve = node.resolve()
    target_addresses = {y.address for y in targets}

    extra = [x for x in ex_resolve.output_concepts if x.address not in target_addresses]

    base = [
        x
        for x in ex_resolve.output_concepts
        if x.address not in {c.address for c in extra}
    ]

    logger.info(
        f"{depth_to_prefix(depth)}{LOGGER_PREFIX} reducing final outputs, "
        f"was {[c.address for c in ex_resolve.output_concepts]} "
        f"with extra {[c.address for c in extra]}, remaining {base}"
    )

    # Add missing targets
    for target in targets:
        if target.address not in {c.address for c in base}:
            base.append(target)

    node.set_output_concepts(base)
    return extra


# Simple factory functions for basic derivation types
def _generate_window_node(ctx: NodeGenerationContext) -> StrategyNode | None:
    ctx.log_generation("window")
    return gen_window_node(
        ctx.concept,
        ctx.local_optional,
        history=ctx.history,
        environment=ctx.environment,
        g=ctx.g,
        depth=ctx.next_depth,
        source_concepts=ctx.source_concepts,
        conditions=ctx.conditions,
    )


def _generate_filter_node(ctx: NodeGenerationContext) -> StrategyNode | None:
    ctx.log_generation("filter")
    return gen_filter_node(
        ctx.concept,
        ctx.local_optional,
        history=ctx.history,
        environment=ctx.environment,
        g=ctx.g,
        depth=ctx.next_depth,
        source_concepts=ctx.source_concepts,
        conditions=ctx.conditions,
    )


def _generate_unnest_node(ctx: NodeGenerationContext) -> StrategyNode | None:
    ctx.log_generation("unnest", f"condition {ctx.conditions}")
    return gen_unnest_node(
        ctx.concept,
        ctx.local_optional,
        history=ctx.history,
        environment=ctx.environment,
        g=ctx.g,
        depth=ctx.next_depth,
        source_concepts=ctx.source_concepts,
        conditions=ctx.conditions,
    )


def _generate_recursive_node(ctx: NodeGenerationContext) -> StrategyNode | None:
    ctx.log_generation("recursive", f"condition {ctx.conditions}")
    return gen_recursive_node(
        ctx.concept,
        ctx.local_optional,
        history=ctx.history,
        environment=ctx.environment,
        g=ctx.g,
        depth=ctx.next_depth,
        source_concepts=ctx.source_concepts,
        conditions=ctx.conditions,
    )


def _generate_union_node(ctx: NodeGenerationContext) -> StrategyNode | None:
    ctx.log_generation("union", f"condition {ctx.conditions}")
    return gen_union_node(
        ctx.concept,
        ctx.local_optional,
        ctx.environment,
        ctx.g,
        ctx.next_depth,
        ctx.source_concepts,
        ctx.history,
        conditions=ctx.conditions,
    )


def _generate_aggregate_node(ctx: NodeGenerationContext) -> StrategyNode | None:
    # Filter out constants to avoid multiplication issues
    agg_optional = [
        x
        for x in ctx.local_optional
        if not (
            x.granularity == Granularity.SINGLE_ROW
            and x.derivation != Derivation.AGGREGATE
        )
    ]

    logger.info(
        f"{depth_to_prefix(ctx.depth)}{LOGGER_PREFIX} "
        f"for {ctx.concept.address}, generating aggregate node with {agg_optional}"
    )

    return gen_group_node(
        ctx.concept,
        agg_optional,
        history=ctx.history,
        environment=ctx.environment,
        g=ctx.g,
        depth=ctx.next_depth,
        source_concepts=ctx.source_concepts,
        conditions=ctx.conditions,
    )


def _generate_rowset_node(ctx: NodeGenerationContext) -> StrategyNode | None:
    ctx.log_generation("rowset")
    return gen_rowset_node(
        ctx.concept,
        ctx.local_optional,
        ctx.environment,
        ctx.g,
        ctx.next_depth,
        ctx.source_concepts,
        ctx.history,
        conditions=ctx.conditions,
    )


def _generate_multiselect_node(ctx: NodeGenerationContext) -> StrategyNode | None:
    ctx.log_generation("multiselect")
    return gen_multiselect_node(
        ctx.concept,
        ctx.local_optional,
        ctx.environment,
        ctx.g,
        ctx.next_depth,
        ctx.source_concepts,
        ctx.history,
        conditions=ctx.conditions,
    )


def _generate_group_to_node(ctx: NodeGenerationContext) -> StrategyNode | None:
    ctx.log_generation("group to grain")
    return gen_group_to_node(
        ctx.concept,
        ctx.local_optional,
        ctx.environment,
        ctx.g,
        ctx.next_depth,
        ctx.source_concepts,
        ctx.history,
        conditions=ctx.conditions,
    )


def _generate_basic_node(ctx: NodeGenerationContext) -> StrategyNode | None:
    ctx.log_generation("basic")
    return gen_basic_node(
        ctx.concept,
        ctx.local_optional,
        history=ctx.history,
        environment=ctx.environment,
        g=ctx.g,
        depth=ctx.next_depth,
        source_concepts=ctx.source_concepts,
        conditions=ctx.conditions,
    )


def _generate_constant_node(ctx: NodeGenerationContext) -> StrategyNode | None:
    ctx.log_generation("constant")
    return gen_constant_node(
        ctx.concept,
        ctx.local_optional,
        history=ctx.history,
        environment=ctx.environment,
        g=ctx.g,
        depth=ctx.next_depth,
        source_concepts=ctx.source_concepts,
        conditions=ctx.conditions,
        accept_partial=ctx.accept_partial,
    )


class RootNodeHandler:
    """Handles complex root node generation logic."""

    def __init__(self, context: NodeGenerationContext):
        self.ctx = context

    def generate(self) -> Optional[StrategyNode]:
        self.ctx.log_generation("select", "including condition inputs")

        root_targets = [self.ctx.concept] + self.ctx.local_optional

        if self._has_non_root_concepts():
            return self._handle_non_root_concepts(root_targets)

        return self._resolve_root_concepts(root_targets)

    def _has_non_root_concepts(self) -> bool:
        return any(
            x.derivation not in (Derivation.ROOT, Derivation.CONSTANT)
            for x in self.ctx.local_optional
        )

    def _handle_non_root_concepts(
        self, root_targets: List[BuildConcept]
    ) -> Optional[StrategyNode]:
        non_root = [
            x.address
            for x in self.ctx.local_optional
            if x.derivation not in (Derivation.ROOT, Derivation.CONSTANT)
        ]

        logger.info(
            f"{depth_to_prefix(self.ctx.depth)}{LOGGER_PREFIX} "
            f"including any filters, there are non-root concepts we should expand first: {non_root}. "
            f"Recursing with all of these as mandatory"
        )

        self.ctx.history.log_start(
            root_targets,
            accept_partial=self.ctx.accept_partial,
            conditions=self.ctx.conditions,
        )

        return self.ctx.source_concepts(
            mandatory_list=root_targets,
            environment=self.ctx.environment,
            g=self.ctx.g,
            depth=self.ctx.next_depth,
            accept_partial=self.ctx.accept_partial,
            history=self.ctx.history,
        )

    def _resolve_root_concepts(
        self, root_targets: List[BuildConcept]
    ) -> Optional[StrategyNode]:
        expanded_node = self._try_merge_expansion(root_targets)
        if expanded_node:
            return expanded_node
        if self.ctx.accept_partial:
            synonym_node = self._try_synonym_resolution(root_targets)
            if synonym_node:
                logger.info(
                    f"{depth_to_prefix(self.ctx.depth)}{LOGGER_PREFIX} "
                    f"resolved root concepts through synonyms"
                )
                return synonym_node

        return None

    def _try_merge_expansion(
        self, root_targets: List[BuildConcept]
    ) -> Optional[StrategyNode]:
        for accept_partial in [False, True]:
            expanded = gen_merge_node(
                all_concepts=root_targets,
                environment=self.ctx.environment,
                g=self.ctx.g,
                depth=self.ctx.next_depth,
                source_concepts=self.ctx.source_concepts,
                history=self.ctx.history,
                search_conditions=self.ctx.conditions,
                accept_partial=accept_partial,
            )

            if expanded:
                self._handle_expanded_node(expanded, root_targets)
                return expanded

        logger.info(
            f"{depth_to_prefix(self.ctx.depth)}{LOGGER_PREFIX} "
            f"could not find additional concept(s) to inject"
        )
        return None

    def _handle_expanded_node(
        self, expanded: StrategyNode, root_targets: List[BuildConcept]
    ) -> None:
        extra = restrict_node_outputs_targets(expanded, root_targets, self.ctx.depth)

        pseudonyms = [
            x
            for x in extra
            if any(x.address in y.pseudonyms for y in root_targets)
            and x.address not in root_targets
        ]

        if pseudonyms:
            expanded.add_output_concepts(pseudonyms)

        logger.info(
            f"{depth_to_prefix(self.ctx.depth)}{LOGGER_PREFIX} "
            f"Found connections for {[c.address for c in root_targets]} "
            f"via concept addition; removing extra {[c.address for c in extra]}"
        )

    def _try_synonym_resolution(
        self, root_targets: List[BuildConcept]
    ) -> Optional[StrategyNode]:
        logger.info(
            f"{depth_to_prefix(self.ctx.depth)}{LOGGER_PREFIX} "
            f"Could not resolve root concepts, checking for synonyms for {root_targets}"
        )

        if not self.ctx.history.check_started(
            root_targets,
            accept_partial=self.ctx.accept_partial,
            conditions=self.ctx.conditions,
        ):
            self.ctx.history.log_start(
                root_targets,
                accept_partial=self.ctx.accept_partial,
                conditions=self.ctx.conditions,
            )

            resolved = gen_synonym_node(
                all_concepts=root_targets,
                environment=self.ctx.environment,
                g=self.ctx.g,
                depth=self.ctx.next_depth,
                source_concepts=self.ctx.source_concepts,
                history=self.ctx.history,
                conditions=self.ctx.conditions,
                accept_partial=self.ctx.accept_partial,
            )

            if resolved:
                logger.info(
                    f"{depth_to_prefix(self.ctx.depth)}{LOGGER_PREFIX} "
                    f"resolved concepts through synonyms"
                )
                return resolved
        else:
            logger.info(
                f"{depth_to_prefix(self.ctx.depth)}{LOGGER_PREFIX} "
                f"skipping synonym search, already in a recursion for these concepts"
            )

        return None


def generate_node(
    concept: BuildConcept,
    local_optional: List[BuildConcept],
    environment: BuildEnvironment,
    g: ReferenceGraph,
    depth: int,
    source_concepts: SearchConceptsType,
    history: History,
    accept_partial: bool = False,
    conditions: BuildWhereClause | None = None,
) -> StrategyNode | None:

    context = NodeGenerationContext(
        concept=concept,
        local_optional=local_optional,
        environment=environment,
        g=g,
        depth=depth,
        source_concepts=source_concepts,
        history=history,
        accept_partial=accept_partial,
        conditions=conditions,
    )

    # Try materialized concept first
    # this is worth checking every loop iteration
    candidate = history.gen_select_node(
        [concept] + local_optional,
        environment,
        g,
        depth + 1,
        fail_if_not_found=False,
        accept_partial=accept_partial,
        conditions=conditions,
    )

    if candidate:
        return candidate

    # Delegate to appropriate handler based on derivation
    derivation_handlers = {
        Derivation.WINDOW: lambda: _generate_window_node(context),
        Derivation.FILTER: lambda: _generate_filter_node(context),
        Derivation.UNNEST: lambda: _generate_unnest_node(context),
        Derivation.RECURSIVE: lambda: _generate_recursive_node(context),
        Derivation.UNION: lambda: _generate_union_node(context),
        Derivation.AGGREGATE: lambda: _generate_aggregate_node(context),
        Derivation.ROWSET: lambda: _generate_rowset_node(context),
        Derivation.MULTISELECT: lambda: _generate_multiselect_node(context),
        Derivation.GROUP_TO: lambda: _generate_group_to_node(context),
        Derivation.BASIC: lambda: _generate_basic_node(context),
        Derivation.ROOT: lambda: RootNodeHandler(context).generate(),
        Derivation.CONSTANT: lambda: _generate_constant_node(context),
    }

    handler = derivation_handlers.get(concept.derivation)
    if not handler:
        raise ValueError(f"Unknown derivation {concept.derivation} on {concept}")

    return handler()



from trilogy import Dialects, Environment
from trilogy.core.query_processor import process_query
from trilogy.render import get_dialect_generator


def test_sql_generators():
    env = Environment()
    _, statements = env.parse("""const a <- 1; select a;""")
    processed = process_query(env, statements[-1])
    for dialect in Dialects:
        generator = get_dialect_generator(dialect)
        compiled = generator.compile_statement(processed)
        assert compiled.startswith("""SELECT\n    :a"""), f"{dialect} compiled"



from functools import singledispatchmethod
from pathlib import Path
from typing import Any, Generator, List, Optional

from sqlalchemy import text

from trilogy.constants import MagicConstants, Rendering, logger
from trilogy.core.enums import FunctionType, Granularity, IOType, ValidationScope
from trilogy.core.models.author import Concept, Function
from trilogy.core.models.build import BuildFunction
from trilogy.core.models.core import ListWrapper, MapWrapper
from trilogy.core.models.datasource import Datasource
from trilogy.core.models.environment import Environment
from trilogy.core.statements.author import (
    ConceptDeclarationStatement,
    CopyStatement,
    ImportStatement,
    MergeStatementV2,
    MultiSelectStatement,
    PersistStatement,
    RawSQLStatement,
    SelectStatement,
    ShowStatement,
    ValidateStatement,
)
from trilogy.core.statements.execute import (
    PROCESSED_STATEMENT_TYPES,
    ProcessedCopyStatement,
    ProcessedQuery,
    ProcessedQueryPersist,
    ProcessedRawSQLStatement,
    ProcessedShowStatement,
    ProcessedValidateStatement,
)
from trilogy.core.validation.common import (
    ValidationTest,
)
from trilogy.dialect.base import BaseDialect
from trilogy.dialect.enums import Dialects
from trilogy.dialect.metadata import (
    generate_result_set,
    handle_concept_declaration,
    handle_datasource,
    handle_import_statement,
    handle_merge_statement,
    handle_processed_show_statement,
    handle_processed_validate_statement,
    handle_show_statement_outputs,
)
from trilogy.engine import ExecutionEngine, ResultProtocol
from trilogy.hooks.base_hook import BaseHook
from trilogy.parser import parse_text
from trilogy.render import get_dialect_generator


class Executor(object):
    def __init__(
        self,
        dialect: Dialects,
        engine: ExecutionEngine,
        environment: Optional[Environment] = None,
        rendering: Rendering | None = None,
        hooks: List[BaseHook] | None = None,
    ):
        self.dialect: Dialects = dialect
        self.engine = engine
        self.environment = environment or Environment()
        self.generator: BaseDialect
        self.logger = logger
        self.hooks = hooks
        self.generator = get_dialect_generator(self.dialect, rendering)
        self.connection = self.engine.connect()
        # TODO: make generic
        if self.dialect == Dialects.DATAFRAME:
            self.engine.setup(self.environment, self.connection)

    def execute_statement(
        self,
        statement: PROCESSED_STATEMENT_TYPES,
    ) -> Optional[ResultProtocol]:
        if not isinstance(statement, PROCESSED_STATEMENT_TYPES):
            return None
        return self.execute_query(statement)

    @singledispatchmethod
    def execute_query(self, query) -> ResultProtocol | None:
        raise NotImplementedError("Cannot execute type {}".format(type(query)))

    @execute_query.register
    def _(self, query: ConceptDeclarationStatement) -> ResultProtocol | None:
        return handle_concept_declaration(query)

    @execute_query.register
    def _(self, query: Datasource) -> ResultProtocol | None:
        return handle_datasource(query)

    @execute_query.register
    def _(self, query: str) -> ResultProtocol | None:
        results = self.execute_text(query)
        if results:
            return results[-1]
        return None

    @execute_query.register
    def _(self, query: SelectStatement) -> ResultProtocol | None:
        sql = self.generator.generate_queries(
            self.environment, [query], hooks=self.hooks
        )
        return self.execute_query(sql[0])

    @execute_query.register
    def _(self, query: PersistStatement) -> ResultProtocol | None:
        sql = self.generator.generate_queries(
            self.environment, [query], hooks=self.hooks
        )
        return self.execute_query(sql[0])

    @execute_query.register
    def _(self, query: RawSQLStatement) -> ResultProtocol | None:
        return self.execute_raw_sql(query.text)

    @execute_query.register
    def _(self, query: ShowStatement) -> ResultProtocol | None:
        sql = self.generator.generate_queries(
            self.environment, [query], hooks=self.hooks
        )
        return self.execute_query(sql[0])

    @execute_query.register
    def _(self, query: ProcessedShowStatement) -> ResultProtocol | None:
        return handle_processed_show_statement(
            query,
            [
                self.generator.compile_statement(x)
                for x in query.output_values
                if isinstance(x, (ProcessedQuery, ProcessedQueryPersist))
            ],
        )

    @execute_query.register
    def _(self, query: ProcessedValidateStatement) -> ResultProtocol | None:
        return handle_processed_validate_statement(
            query, self.generator, self.validate_environment
        )

    @execute_query.register
    def _(self, query: ImportStatement) -> ResultProtocol | None:
        return handle_import_statement(query)

    @execute_query.register
    def _(self, query: MergeStatementV2) -> ResultProtocol | None:
        return handle_merge_statement(query, self.environment)

    @execute_query.register
    def _(self, query: ProcessedRawSQLStatement) -> ResultProtocol | None:
        return self.execute_raw_sql(query.text)

    @execute_query.register
    def _(self, query: ProcessedQuery) -> ResultProtocol | None:
        sql = self.generator.compile_statement(query)
        output = self.execute_raw_sql(sql, local_concepts=query.local_concepts)
        return output

    @execute_query.register
    def _(self, query: ProcessedQueryPersist) -> ResultProtocol | None:
        sql = self.generator.compile_statement(query)

        output = self.execute_raw_sql(sql, local_concepts=query.local_concepts)
        self.environment.add_datasource(query.datasource)
        return output

    @execute_query.register
    def _(self, query: ProcessedCopyStatement) -> ResultProtocol | None:
        sql = self.generator.compile_statement(query)
        output: ResultProtocol = self.execute_raw_sql(
            sql, local_concepts=query.local_concepts
        )
        if query.target_type == IOType.CSV:
            import csv

            with open(query.target, "w", newline="", encoding="utf-8") as f:
                outcsv = csv.writer(f)
                outcsv.writerow(output.keys())
                outcsv.writerows(output)
        else:
            raise NotImplementedError(f"Unsupported IO Type {query.target_type}")
        # now return the query we ran through IO
        # TODO: instead return how many rows were written?
        return generate_result_set(
            query.output_columns,
            [self.generator.compile_statement(query)],
        )

    @singledispatchmethod
    def generate_sql(self, command) -> list[str]:
        raise NotImplementedError(
            "Cannot generate sql for type {}".format(type(command))
        )

    @generate_sql.register  # type: ignore
    def _(self, command: ProcessedQuery) -> List[str]:
        output = []
        compiled_sql = self.generator.compile_statement(command)
        output.append(compiled_sql)
        return output

    @generate_sql.register
    def _(self, command: ProcessedShowStatement) -> List[str]:
        output = []
        for statement in command.output_values:
            if isinstance(statement, (ProcessedQuery, ProcessedQueryPersist)):
                compiled_sql = self.generator.compile_statement(statement)
                output.append(compiled_sql)
        return output

    @generate_sql.register  # type: ignore
    def _(self, command: MultiSelectStatement) -> List[str]:
        output = []
        sql = self.generator.generate_queries(
            self.environment, [command], hooks=self.hooks
        )
        for statement in sql:
            compiled_sql = self.generator.compile_statement(statement)
            output.append(compiled_sql)
        return output

    @generate_sql.register  # type: ignore
    def _(self, command: SelectStatement) -> List[str]:
        output = []
        sql = self.generator.generate_queries(
            self.environment, [command], hooks=self.hooks
        )
        for statement in sql:
            compiled_sql = self.generator.compile_statement(statement)
            output.append(compiled_sql)
        return output

    @generate_sql.register  # type: ignore
    def _(self, command: str) -> List[str]:
        """generate SQL for execution"""
        _, parsed = parse_text(command, self.environment)
        generatable = [
            x for x in parsed if isinstance(x, (SelectStatement, PersistStatement))
        ]
        sql = self.generator.generate_queries(
            self.environment, generatable, hooks=self.hooks
        )
        output = []
        for statement in sql:
            if isinstance(statement, ProcessedShowStatement):
                continue
            compiled_sql = self.generator.compile_statement(statement)
            output.append(compiled_sql)
        return output

    def parse_file(
        self, file: str | Path, persist: bool = False
    ) -> list[PROCESSED_STATEMENT_TYPES]:
        return list(self.parse_file_generator(file, persist=persist))

    def parse_file_generator(
        self, file: str | Path, persist: bool = False
    ) -> Generator[
        PROCESSED_STATEMENT_TYPES,
        None,
        None,
    ]:
        file = Path(file)
        with open(file, "r") as f:
            command = f.read()
            return self.parse_text_generator(command, persist=persist, root=file)

    def parse_text(
        self, command: str, persist: bool = False, root: Path | None = None
    ) -> List[PROCESSED_STATEMENT_TYPES]:
        return list(self.parse_text_generator(command, persist=persist, root=root))

    def parse_text_generator(
        self, command: str, persist: bool = False, root: Path | None = None
    ) -> Generator[
        PROCESSED_STATEMENT_TYPES,
        None,
        None,
    ]:
        """Process a preql text command"""
        _, parsed = parse_text(command, self.environment, root=root)
        generatable = [
            x
            for x in parsed
            if isinstance(
                x,
                (
                    SelectStatement,
                    PersistStatement,
                    MultiSelectStatement,
                    ShowStatement,
                    RawSQLStatement,
                    CopyStatement,
                    ValidateStatement,
                ),
            )
        ]
        while generatable:
            t = generatable.pop(0)
            x = self.generator.generate_queries(
                self.environment, [t], hooks=self.hooks
            )[0]

            yield x

            if persist and isinstance(x, ProcessedQueryPersist):
                self.environment.add_datasource(x.datasource)

    def _atom_to_value(self, val: Any) -> Any:
        if val == MagicConstants.NULL:
            return None
        return val

    def _concept_to_value(
        self,
        concept: Concept,
        local_concepts: dict[str, Concept] | None = None,
    ) -> Any:
        if not concept.granularity == Granularity.SINGLE_ROW:
            raise SyntaxError(f"Cannot bind non-singleton concept {concept.address}")
        # TODO: to get rid of function here - need to figure out why it's getting passed in
        if (
            isinstance(concept.lineage, (BuildFunction, Function))
            and concept.lineage.operator == FunctionType.CONSTANT
        ):
            rval = concept.lineage.arguments[0]
            if isinstance(rval, ListWrapper):
                return [self._atom_to_value(x) for x in rval]
            if isinstance(rval, MapWrapper):
                # duckdb expects maps in this format as variables
                if self.dialect == Dialects.DUCK_DB:
                    return {
                        "key": [self._atom_to_value(x) for x in rval],
                        "value": [self._atom_to_value(rval[x]) for x in rval],
                    }
                return {k: self._atom_to_value(v) for k, v in rval.items()}
            # if isinstance(rval, ConceptRef):
            #     return self._concept_to_value(self.environment.concepts[rval.address], local_concepts=local_concepts)
            return rval
        else:
            results = self.execute_query(f"select {concept.name} limit 1;")
            if results:
                fetcher = results.fetchone()
                if fetcher:
                    return fetcher[0]
            return None

    def _hydrate_param(
        self, param: str, local_concepts: dict[str, Concept] | None = None
    ) -> Any:
        matched = [
            v
            for v in self.environment.concepts.values()
            if v.safe_address == param or v.address == param
        ]
        if local_concepts and not matched:
            matched = [
                v
                for v in local_concepts.values()
                if v.safe_address == param or v.address == param
            ]
        if not matched:
            raise SyntaxError(f"No concept found for parameter {param}")

        concept: Concept = matched.pop()
        return self._concept_to_value(concept, local_concepts=local_concepts)

    def execute_raw_sql(
        self,
        command: str | Path,
        variables: dict | None = None,
        local_concepts: dict[str, Concept] | None = None,
    ) -> ResultProtocol:
        """Run a command against the raw underlying
        execution engine."""
        final_params = None
        if isinstance(command, Path):
            with open(command, "r") as f:
                command = f.read()
        q = text(command)
        if variables:
            final_params = variables
        else:
            params = q.compile().params
            if params:
                final_params = {
                    x: self._hydrate_param(x, local_concepts=local_concepts)
                    for x in params
                }

        if final_params:
            return self.connection.execute(text(command), final_params)
        return self.connection.execute(
            text(command),
        )

    def execute_text(
        self, command: str, non_interactive: bool = False
    ) -> List[ResultProtocol]:
        """Run a trilogy query expressed as text."""
        output: list[ResultProtocol] = []
        # connection = self.engine.connect()
        for statement in self.parse_text_generator(command):
            if isinstance(statement, ProcessedShowStatement):
                results = handle_show_statement_outputs(
                    statement,
                    [
                        self.generator.compile_statement(x)
                        for x in statement.output_values
                        if isinstance(x, (ProcessedQuery, ProcessedQueryPersist))
                    ],
                    self.environment,
                    self.generator,
                )
                output.extend(results)
                continue
            elif isinstance(statement, ProcessedValidateStatement):
                validate_result = handle_processed_validate_statement(
                    statement, self.generator, self.validate_environment
                )
                if validate_result:
                    output.append(validate_result)
                continue
            if non_interactive:
                if not isinstance(
                    statement, (ProcessedCopyStatement, ProcessedQueryPersist)
                ):
                    continue
            result = self.execute_statement(statement)
            if result:
                output.append(result)
        return output

    def execute_file(
        self, file: str | Path, non_interactive: bool = False
    ) -> List[ResultProtocol]:
        file = Path(file)
        with open(file, "r") as f:
            command = f.read()
        return self.execute_text(command, non_interactive=non_interactive)

    def validate_environment(
        self,
        scope: ValidationScope = ValidationScope.ALL,
        targets: Optional[List[str]] = None,
        generate_only: bool = False,
    ) -> list[ValidationTest]:
        from trilogy.core.validation.environment import validate_environment

        return validate_environment(
            self.environment, scope, targets, exec=None if generate_only else self
        )



from typing import List

from trilogy.constants import logger
from trilogy.core.enums import FunctionClass, FunctionType, SourceType
from trilogy.core.models.build import BuildConcept, BuildFunction, BuildWhereClause
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.processing.node_generators.common import (
    resolve_function_parent_concepts,
)
from trilogy.core.processing.nodes import History, StrategyNode
from trilogy.utility import unique

LOGGER_PREFIX = "[GEN_BASIC_NODE]"


def is_equivalent_basic_function_lineage(
    x: BuildConcept,
    y: BuildConcept,
):
    if not isinstance(x.lineage, BuildFunction) or not isinstance(
        y.lineage, BuildFunction
    ):
        return False
    if x.lineage.operator == y.lineage.operator == FunctionType.ATTR_ACCESS:
        return x.lineage.concept_arguments == y.lineage.concept_arguments
    if x.lineage.operator == y.lineage.operator:
        return True
    if (
        y.lineage.operator in FunctionClass.AGGREGATE_FUNCTIONS.value
        or y.lineage.operator in FunctionClass.ONE_TO_MANY.value
    ):
        return False
    return True


def gen_basic_node(
    concept: BuildConcept,
    local_optional: List[BuildConcept],
    environment: BuildEnvironment,
    g,
    depth: int,
    source_concepts,
    history: History | None = None,
    conditions: BuildWhereClause | None = None,
):
    depth_prefix = "\t" * depth
    parent_concepts = resolve_function_parent_concepts(concept, environment=environment)

    logger.info(
        f"{depth_prefix}{LOGGER_PREFIX} basic node for {concept} with lineage {concept.lineage} has parents {[x for x in parent_concepts]}"
    )
    synonyms: list[BuildConcept] = []
    ignored_optional: set[str] = set()
    assert isinstance(concept.lineage, BuildFunction)
    # when we are getting an attribute, if there is anything else
    # that is an attribute of the same struct in local optional
    # select that value for discovery as well
    if concept.lineage.operator == FunctionType.ATTR_ACCESS:
        logger.info(
            f"{depth_prefix}{LOGGER_PREFIX} checking for synonyms for attribute access"
        )
        for x in local_optional:
            found = False
            for z in x.pseudonyms:
                # gate to ensure we don't match to multiple synonyms
                if found:
                    continue
                if z in environment.concepts:
                    s_concept = environment.concepts[z]
                else:
                    s_concept = environment.alias_origin_lookup[z]
                if is_equivalent_basic_function_lineage(concept, s_concept):
                    found = True
                    synonyms.append(s_concept)
                    ignored_optional.add(x.address)
    equivalent_optional = [
        x
        for x in local_optional
        if is_equivalent_basic_function_lineage(concept, x)
        and x.address != concept.address
    ] + synonyms

    if equivalent_optional:
        logger.info(
            f"{depth_prefix}{LOGGER_PREFIX} basic node for {concept} has equivalent optional {[x.address for x in equivalent_optional]}"
        )
    for eo in equivalent_optional:
        new_parents = resolve_function_parent_concepts(eo, environment=environment)
        logger.info(
            f"{depth_prefix}{LOGGER_PREFIX} equivalent optional {eo.address} has parents {[x.address for x in new_parents]}"
        )
        parent_concepts += new_parents
    non_equivalent_optional = [
        x
        for x in local_optional
        if x not in equivalent_optional
        and not any(x.address in y.pseudonyms for y in equivalent_optional)
        and x.address not in ignored_optional
    ]
    logger.info(
        f"{depth_prefix}{LOGGER_PREFIX} basic node for {concept} has non-equivalent optional {[x.address for x in non_equivalent_optional]}"
    )
    all_parents: list[BuildConcept] = unique(
        parent_concepts + non_equivalent_optional, "address"
    )
    logger.info(
        f"{depth_prefix}{LOGGER_PREFIX} Fetching parents {[x.address for x in all_parents]}"
    )
    parent_node: StrategyNode | None = source_concepts(
        mandatory_list=all_parents,
        environment=environment,
        g=g,
        depth=depth + 1,
        history=history,
        conditions=conditions,
    )

    if not parent_node:
        logger.info(
            f"{depth_prefix}{LOGGER_PREFIX} No basic node could be generated for {concept}"
        )
        return None
    if parent_node.source_type != SourceType.CONSTANT:
        parent_node.source_type = SourceType.BASIC
    parent_node.add_output_concept(concept)
    for x in equivalent_optional:
        parent_node.add_output_concept(x)

    logger.info(
        f"{depth_prefix}{LOGGER_PREFIX} Returning basic select for {concept}: output {[x.address for x in parent_node.output_concepts]}"
    )
    # if it's a constant, don't prune outputs
    if parent_node.source_type == SourceType.CONSTANT:
        return parent_node
    targets = [concept] + local_optional + equivalent_optional
    targets = [
        s
        for s in parent_node.output_concepts
        if any(s.address in y.pseudonyms for y in targets)
    ] + targets
    parent_node.set_output_concepts(targets)

    logger.info(
        f"{depth_prefix}{LOGGER_PREFIX} Returning basic select for {concept}: input: {[x.address for x in parent_node.input_concepts]} output {[x.address for x in parent_node.output_concepts]} hidden {[x for x in parent_node.hidden_concepts]}"
    )

    return parent_node



from dataclasses import dataclass
from enum import Enum

from trilogy import Environment
from trilogy.authoring import (
    ConceptRef,
    DataType,
    Ordering,
    Purpose,
)
from trilogy.constants import MagicConstants
from trilogy.core.enums import ComparisonOperator, FunctionType
from trilogy.core.exceptions import ModelValidationError
from trilogy.core.models.build import (
    BuildCaseElse,
    BuildCaseWhen,
    BuildComparison,
    BuildConcept,
    BuildConditional,
    BuildDatasource,
    BuildFunction,
    BuildOrderBy,
    BuildOrderItem,
)
from trilogy.core.models.environment import EnvironmentConceptDict
from trilogy.core.models.execute import (
    CTE,
    QueryDatasource,
)
from trilogy.core.statements.execute import ProcessedQuery


class ExpectationType(Enum):
    LOGICAL = "logical"
    ROWCOUNT = "rowcount"
    DATA_TYPE_LIST = "data_type_list"


@dataclass
class ValidationTest:
    check_type: ExpectationType
    raw_query: ProcessedQuery | None = None
    generated_query: str | None = None
    expected: str | None = None
    result: ModelValidationError | None = None
    ran: bool = True


class ValidationType(Enum):
    DATASOURCES = "datasources"
    CONCEPTS = "concepts"


def build_order_args(concepts: list[BuildConcept]) -> list[BuildFunction]:
    order_args = []
    for concept in concepts:
        order_args.append(
            BuildFunction(
                operator=FunctionType.CASE,
                arguments=[
                    BuildCaseWhen(
                        comparison=BuildComparison(
                            left=concept,
                            operator=ComparisonOperator.IS,
                            right=MagicConstants.NULL,
                        ),
                        expr=1,
                    ),
                    BuildCaseElse(expr=0),
                ],
                output_data_type=DataType.INTEGER,
                output_purpose=Purpose.PROPERTY,
                arg_count=2,
            )
        )

    return order_args


def easy_query(
    concepts: list[BuildConcept],
    datasource: BuildDatasource,
    env: Environment,
    condition: BuildConditional | BuildComparison | None = None,
    limit: int = 100,
):
    """
    Build basic datasource specific queries.
    """
    datasource_outputs = {c.address: c for c in datasource.concepts}
    first_qds_concepts = datasource.concepts + concepts
    root_qds = QueryDatasource(
        input_concepts=first_qds_concepts,
        output_concepts=concepts,
        datasources=[datasource],
        joins=[],
        source_map={
            concept.address: (
                set([datasource]) if concept.address in datasource_outputs else set()
            )
            # include all base datasource conepts for convenience
            for concept in first_qds_concepts
        },
        grain=datasource.grain,
    )
    cte = CTE(
        name=f"datasource_{datasource.name}_base",
        source=root_qds,
        output_columns=concepts,
        source_map={
            concept.address: (
                [datasource.safe_identifier]
                if concept.address in datasource_outputs
                else []
            )
            for concept in first_qds_concepts
        },
        grain=datasource.grain,
        group_to_grain=True,
        base_alias_override=datasource.safe_identifier,
    )
    filter_cte = CTE(
        name=f"datasource_{datasource.name}_filter",
        source=QueryDatasource(
            datasources=[root_qds],
            input_concepts=cte.output_columns,
            output_concepts=cte.output_columns,
            joins=[],
            source_map={concept.address: (set([root_qds])) for concept in concepts},
            grain=cte.grain,
        ),
        parent_ctes=[cte],
        output_columns=cte.output_columns,
        source_map={
            concept.address: [cte.identifier] for concept in cte.output_columns
        },
        grain=cte.grain,
        condition=condition,
        limit=limit,
        order_by=BuildOrderBy(
            items=[
                BuildOrderItem(
                    expr=BuildFunction(
                        operator=FunctionType.SUM,
                        arguments=build_order_args(concepts),
                        output_data_type=DataType.INTEGER,
                        output_purpose=Purpose.PROPERTY,
                        arg_count=len(concepts),
                    ),
                    order=Ordering.DESCENDING,
                )
            ]
        ),
    )

    return ProcessedQuery(
        output_columns=[ConceptRef(address=concept.address) for concept in concepts],
        ctes=[cte, filter_cte],
        base=cte,
        local_concepts=EnvironmentConceptDict(**{}),
    )



from functools import reduce
from typing import TYPE_CHECKING, List, Optional

import networkx as nx

from trilogy.constants import logger
from trilogy.core.enums import Derivation
from trilogy.core.graph_models import (
    ReferenceGraph,
    concept_to_node,
    get_graph_exact_match,
    prune_sources_for_conditions,
)
from trilogy.core.models.build import (
    BuildConcept,
    BuildDatasource,
    BuildGrain,
    BuildWhereClause,
    LooseBuildConceptList,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.processing.node_generators.select_helpers.datasource_injection import (
    get_union_sources,
)
from trilogy.core.processing.nodes import (
    ConstantNode,
    GroupNode,
    MergeNode,
    SelectNode,
    StrategyNode,
)
from trilogy.core.processing.utility import padding

if TYPE_CHECKING:
    from trilogy.core.processing.nodes.union_node import UnionNode

LOGGER_PREFIX = "[GEN_ROOT_MERGE_NODE]"


def extract_address(node: str):
    return node.split("~")[1].split("@")[0]


def get_graph_partial_nodes(
    g: ReferenceGraph, conditions: BuildWhereClause | None
) -> dict[str, list[str]]:
    partial: dict[str, list[str]] = {}
    for node, ds in g.datasources.items():

        if not isinstance(ds, list):

            if ds.non_partial_for and conditions == ds.non_partial_for:
                partial[node] = []
                continue
            partial[node] = [concept_to_node(c) for c in ds.partial_concepts]
        # assume union sources have no partial
        else:
            partial[node] = []
    return partial


def get_graph_grains(g: ReferenceGraph) -> dict[str, list[str]]:
    grain_length: dict[str, list[str]] = {}
    for node, lookup in g.datasources.items():

        base: set[str] = set()
        if not isinstance(lookup, list):
            flookup = [lookup]
        else:
            flookup = lookup
        assert isinstance(flookup, list)
        grain_length[node] = reduce(
            lambda x, y: x.union(y.grain.components), flookup, base  # type: ignore
        )
    return grain_length


def subgraph_is_complete(
    nodes: list[str], targets: set[str], mapping: dict[str, str], g: nx.DiGraph
) -> bool:
    # Check if all targets are present in mapped nodes
    mapped = {mapping.get(n, n) for n in nodes}
    if not targets.issubset(mapped):
        # logger.info(
        #     f"Subgraph {nodes} is not complete, missing targets {targets} - mapped {mapped}"
        # )
        return False

    # Check if at least one concept node has a datasource edge
    has_ds_edge = {target: False for target in targets}

    for node in nodes:
        if node.startswith("c~"):
            mapped_node = mapping.get(node, node)
            if mapped_node in targets and not has_ds_edge[mapped_node]:
                # Only check neighbors if we haven't found a ds edge for this mapped node yet
                if any(
                    neighbor.startswith("ds~") for neighbor in nx.neighbors(g, node)
                ):
                    has_ds_edge[mapped_node] = True

    return all(has_ds_edge.values())


def create_pruned_concept_graph(
    g: ReferenceGraph,
    all_concepts: List[BuildConcept],
    datasources: list[BuildDatasource],
    accept_partial: bool = False,
    conditions: BuildWhereClause | None = None,
    depth: int = 0,
) -> nx.DiGraph:
    orig_g = g

    g = g.copy()
    union_options = get_union_sources(datasources, all_concepts)

    for ds_list in union_options:
        node_address = "ds~" + "-".join([x.name for x in ds_list])
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} injecting potentially relevant union datasource {node_address}"
        )
        common: set[BuildConcept] = set.intersection(
            *[set(x.output_concepts) for x in ds_list]
        )
        g.add_datasource_node(node_address, ds_list)
        for c in common:
            cnode = concept_to_node(c)
            g.add_edge(node_address, cnode)
            g.add_edge(cnode, node_address)
    prune_sources_for_conditions(g, accept_partial, conditions)
    target_addresses = set([c.address for c in all_concepts])
    concepts: dict[str, BuildConcept] = orig_g.concepts
    datasource_map: dict[str, BuildDatasource] = orig_g.datasources
    relevant_concepts_pre = {
        n: x.address
        for n in g.nodes()
        # filter out synonyms
        if (x := concepts.get(n, None)) and x.address in target_addresses
    }

    relevant_concepts: list[str] = list(relevant_concepts_pre.keys())
    relevent_datasets: list[str] = []
    if not accept_partial:
        partial = {}
        partial = get_graph_partial_nodes(g, conditions)
        to_remove = []
        for edge in g.edges:
            if (
                edge[0] in datasource_map
                and (pnodes := partial.get(edge[0], []))
                and edge[1] in pnodes
            ):
                to_remove.append(edge)
            if (
                edge[1] in datasource_map
                and (pnodes := partial.get(edge[1], []))
                and edge[0] in pnodes
            ):
                to_remove.append(edge)
        for edge in to_remove:
            g.remove_edge(*edge)

    for n in g.datasources:
        if any([[n, x] in g.edges for x in relevant_concepts]):
            relevent_datasets.append(n)
            continue

    # for injecting extra join concepts that are shared between datasets
    # use the original graph, pre-partial pruning
    for n in orig_g.concepts:
        # readd ignoring grain
        # we want to join inclusive of all concepts
        if n not in relevant_concepts:
            n_neighbors = nx.all_neighbors(orig_g, n)
            # check if the irrelevant concept is a join between
            # two relevant datasets
            neighbors = set()
            for neighbor in n_neighbors:
                if neighbor in relevent_datasets:
                    neighbors.add(neighbor)
                    if len(neighbors) > 1:
                        relevant_concepts.append(n)
                        continue
    g.remove_nodes_from(
        [
            n
            for n in g.nodes()
            if n not in relevent_datasets and n not in relevant_concepts
        ]
    )
    # from trilogy.hooks.graph_hook import GraphHook
    # GraphHook().query_graph_built(g)
    subgraphs = list(nx.connected_components(g.to_undirected()))
    subgraphs = [
        s
        for s in subgraphs
        if subgraph_is_complete(s, target_addresses, relevant_concepts_pre, g)
    ]
    # from trilogy.hooks.graph_hook import GraphHook
    # GraphHook().query_graph_built(g)
    if not subgraphs:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} cannot resolve root graph - no subgraphs after node prune"
        )
        return None

    if subgraphs and len(subgraphs) != 1:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} cannot resolve root graph - subgraphs are split - have {len(subgraphs)} from {subgraphs}"
        )
        return None
    # add back any relevant edges that might have been partially filtered
    relevant = set(relevant_concepts + relevent_datasets)
    for edge in orig_g.edges():
        if edge[0] in relevant and edge[1] in relevant:
            g.add_edge(edge[0], edge[1], fast=True)
    # if we have no ds nodes at all, for non constant, we can't find it
    if not any([n.startswith("ds~") for n in g.nodes]):
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} cannot resolve root graph - No datasource nodes found"
        )
        return None
    return g


def resolve_subgraphs(
    g: ReferenceGraph,
    relevant: list[BuildConcept],
    accept_partial: bool,
    conditions: BuildWhereClause | None,
    depth: int = 0,
) -> dict[str, list[str]]:
    """When we have multiple distinct subgraphs within our matched
    nodes that can satisfy a query, resolve which one of those we should
    ultimately ues.
    This should generally return one subgraph for each
    unique set of sub concepts that can be referenced,
    discarding duplicates.
    Duplicate subgraphs will be resolved based on which
    ones are most 'optimal' to use, a hueristic
    that can evolve in the future but is currently based on datasource
    cardinality."""
    datasources = [n for n in g.nodes if n.startswith("ds~")]
    subgraphs: dict[str, list[str]] = {
        ds: list(set(list(nx.all_neighbors(g, ds)))) for ds in datasources
    }
    partial_map = get_graph_partial_nodes(g, conditions)
    exact_map = get_graph_exact_match(g, accept_partial, conditions)
    grain_length = get_graph_grains(g)
    concepts: dict[str, BuildConcept] = g.concepts
    non_partial_map = {
        ds: [concepts[c].address for c in subgraphs[ds] if c not in partial_map[ds]]
        for ds in datasources
    }
    concept_map = {
        ds: [concepts[c].address for c in subgraphs[ds]] for ds in datasources
    }
    pruned_subgraphs = {}

    def score_node(input: str):
        logger.debug(f"{padding(depth)}{LOGGER_PREFIX} scoring node {input}")
        grain = grain_length[input]
        # first - go for lowest grain
        # but if the object we want is in the grain, treat that as "free"
        # ex - pick source with grain(product_id) over grain(order_id)
        # when going for product_id
        score = (
            len(list(grain)) - sum([1 for x in concept_map[input] if x in grain]),
            # then check if it's an exact condition match
            0 if input in exact_map else 0.5,
            # last, number of concepts
            len(subgraphs[input]),
            input,
        )
        logger.debug(f"{padding(depth)}{LOGGER_PREFIX} node {input} has score {score}")
        return score

    for key, nodes in subgraphs.items():

        value = non_partial_map[key]
        all_concepts = concept_map[key]
        is_subset = False
        matches = set()
        # Compare current list with other lists
        for other_key, other_all_concepts in concept_map.items():
            other_value = non_partial_map[other_key]
            # needs to be a subset of non partial and a subset of all
            if (
                key != other_key
                and set(value).issubset(set(other_value))
                and set(all_concepts).issubset(set(other_all_concepts))
            ):
                if len(value) < len(other_value):
                    is_subset = True
                    logger.debug(
                        f"{padding(depth)}{LOGGER_PREFIX} Dropping subgraph {key} with {value} as it is a subset of {other_key} with {other_value}"
                    )
                elif len(value) == len(other_value) and len(all_concepts) == len(
                    other_all_concepts
                ):
                    matches.add(other_key)
                    matches.add(key)
        if matches and not is_subset:
            min_node = min(matches, key=score_node)
            logger.debug(
                f"{padding(depth)}{LOGGER_PREFIX} minimum source score is {min_node}"
            )
            is_subset = key is not min(matches, key=score_node)
        if not is_subset:
            pruned_subgraphs[key] = nodes

    final_nodes: set[str] = set([n for v in pruned_subgraphs.values() for n in v])
    relevant_concepts_pre = {
        n: x.address
        for n in g.nodes()
        # filter out synonyms
        if (x := concepts.get(n, None)) and x.address in relevant
    }
    for node in final_nodes:
        keep = True
        if node.startswith("c~") and node not in relevant_concepts_pre:
            keep = (
                sum(
                    [
                        1 if node in sub_nodes else 0
                        for _, sub_nodes in pruned_subgraphs.items()
                    ]
                )
                > 1
            )
        if not keep:
            logger.debug(
                f"{padding(depth)}{LOGGER_PREFIX} Pruning node {node} as irrelevant after subgraph resolution"
            )
            pruned_subgraphs = {
                k: [n for n in v if n != node] for k, v in pruned_subgraphs.items()
            }

    return pruned_subgraphs


def create_datasource_node(
    datasource: BuildDatasource,
    all_concepts: List[BuildConcept],
    accept_partial: bool,
    environment: BuildEnvironment,
    depth: int,
    conditions: BuildWhereClause | None = None,
) -> tuple[StrategyNode, bool]:

    target_grain = BuildGrain.from_concepts(all_concepts, environment=environment)
    force_group = False
    if not datasource.grain.issubset(target_grain):
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX}_DS_NODE Select node must be wrapped in group, {datasource.grain} not subset of target grain {target_grain}"
        )
        force_group = True
    else:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX}_DS_NODE Select node grain {datasource.grain} is subset of target grain {target_grain}, no group required"
        )
    if not datasource.grain.components:
        force_group = True
    partial_concepts = [
        c.concept
        for c in datasource.columns
        if not c.is_complete and c.concept.address in all_concepts
    ]

    partial_lcl = LooseBuildConceptList(concepts=partial_concepts)
    nullable_concepts = [
        c.concept
        for c in datasource.columns
        if c.is_nullable and c.concept.address in all_concepts
    ]

    nullable_lcl = LooseBuildConceptList(concepts=nullable_concepts)
    partial_is_full = conditions and (conditions == datasource.non_partial_for)

    datasource_conditions = datasource.where.conditional if datasource.where else None
    rval = SelectNode(
        input_concepts=[c.concept for c in datasource.columns],
        output_concepts=sorted(all_concepts, key=lambda x: x.address),
        environment=environment,
        parents=[],
        depth=depth,
        partial_concepts=(
            [] if partial_is_full else [c for c in all_concepts if c in partial_lcl]
        ),
        nullable_concepts=[c for c in all_concepts if c in nullable_lcl],
        accept_partial=accept_partial,
        datasource=datasource,
        grain=datasource.grain,
        conditions=datasource_conditions,
        preexisting_conditions=(
            conditions.conditional if partial_is_full and conditions else None
        ),
    )
    return (
        rval,
        force_group,
    )


def create_union_datasource(
    datasource: list[BuildDatasource],
    all_concepts: List[BuildConcept],
    accept_partial: bool,
    environment: BuildEnvironment,
    depth: int,
    conditions: BuildWhereClause | None = None,
) -> tuple["UnionNode", bool]:
    from trilogy.core.processing.nodes.union_node import UnionNode

    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} generating union node parents with condition {conditions}"
    )
    force_group = False
    parents = []
    for x in datasource:
        subnode, fg = create_datasource_node(
            x,
            all_concepts,
            accept_partial,
            environment,
            depth + 1,
            conditions=conditions,
        )
        parents.append(subnode)
        force_group = force_group or fg
    logger.info(f"{padding(depth)}{LOGGER_PREFIX} returning union node")
    return (
        UnionNode(
            output_concepts=all_concepts,
            input_concepts=all_concepts,
            environment=environment,
            parents=parents,
            depth=depth,
            partial_concepts=[],
        ),
        force_group,
    )


def create_select_node(
    ds_name: str,
    subgraph: list[str],
    accept_partial: bool,
    g: ReferenceGraph,
    environment: BuildEnvironment,
    depth: int,
    conditions: BuildWhereClause | None = None,
) -> StrategyNode:

    all_concepts = [
        environment.concepts[extract_address(c)] for c in subgraph if c.startswith("c~")
    ]

    if all([c.derivation == Derivation.CONSTANT for c in all_concepts]):
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} All concepts {[x.address for x in all_concepts]} are constants, returning constant node"
        )
        return ConstantNode(
            output_concepts=all_concepts,
            input_concepts=[],
            environment=environment,
            parents=[],
            depth=depth,
            # no partial for constants
            partial_concepts=[],
            force_group=False,
            preexisting_conditions=conditions.conditional if conditions else None,
        )

    datasource: BuildDatasource = g.datasources[ds_name]
    if isinstance(datasource, BuildDatasource):
        bcandidate, force_group = create_datasource_node(
            datasource,
            all_concepts,
            accept_partial,
            environment,
            depth,
            conditions=conditions,
        )

    elif isinstance(datasource, list):
        bcandidate, force_group = create_union_datasource(
            datasource,
            all_concepts,
            accept_partial,
            environment,
            depth,
            conditions=conditions,
        )
    else:
        raise ValueError(f"Unknown datasource type {datasource}")

    # we need to nest the group node one further
    if force_group is True:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} source requires group before consumption."
        )
        candidate: StrategyNode = GroupNode(
            output_concepts=all_concepts,
            input_concepts=all_concepts,
            environment=environment,
            parents=[bcandidate],
            depth=depth + 1,
            partial_concepts=bcandidate.partial_concepts,
            nullable_concepts=bcandidate.nullable_concepts,
            preexisting_conditions=bcandidate.preexisting_conditions,
            force_group=force_group,
        )
    else:

        candidate = bcandidate

    return candidate


def gen_select_merge_node(
    all_concepts: List[BuildConcept],
    g: nx.DiGraph,
    environment: BuildEnvironment,
    depth: int,
    accept_partial: bool = False,
    conditions: BuildWhereClause | None = None,
) -> Optional[StrategyNode]:
    non_constant = [c for c in all_concepts if c.derivation != Derivation.CONSTANT]
    constants = [c for c in all_concepts if c.derivation == Derivation.CONSTANT]
    if not non_constant and constants:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} only constant inputs to discovery ({constants}), returning constant node directly"
        )
        for x in constants:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} {x} {x.lineage} {x.derivation}"
            )
        if conditions:
            if not all(
                [x.derivation == Derivation.CONSTANT for x in conditions.row_arguments]
            ):
                logger.info(
                    f"{padding(depth)}{LOGGER_PREFIX} conditions being passed in to constant node {conditions}, but not all concepts are constants."
                )
                return None
            else:
                constants += conditions.row_arguments

        return ConstantNode(
            output_concepts=constants,
            input_concepts=[],
            environment=environment,
            parents=[],
            depth=depth,
            partial_concepts=[],
            force_group=False,
            conditions=conditions.conditional if conditions else None,
        )
    attempts = [
        False,
    ]
    if accept_partial:
        attempts.append(True)
    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} searching for root source graph for concepts {[c.address for c in all_concepts]} and conditions {conditions}"
    )
    pruned_concept_graph = None
    for attempt in attempts:
        pruned_concept_graph = create_pruned_concept_graph(
            g,
            non_constant,
            accept_partial=attempt,
            conditions=conditions,
            datasources=list([x for x in environment.datasources.values()]),
            depth=depth,
        )
        if pruned_concept_graph:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} found covering graph w/ partial flag {attempt}"
            )
            break

    if not pruned_concept_graph:
        logger.info(f"{padding(depth)}{LOGGER_PREFIX} no covering graph found.")
        return None

    sub_nodes = resolve_subgraphs(
        pruned_concept_graph,
        relevant=non_constant,
        accept_partial=accept_partial,
        conditions=conditions,
        depth=depth,
    )

    logger.info(f"{padding(depth)}{LOGGER_PREFIX} fetching subgraphs {sub_nodes}")

    parents = [
        create_select_node(
            k,
            subgraph,
            g=pruned_concept_graph,
            accept_partial=accept_partial,
            environment=environment,
            depth=depth,
            conditions=conditions,
        )
        for k, subgraph in sub_nodes.items()
    ]
    if not parents:
        return None

    if constants:
        parents.append(
            ConstantNode(
                output_concepts=constants,
                input_concepts=[],
                environment=environment,
                parents=[],
                depth=depth,
                partial_concepts=[],
                force_group=False,
                preexisting_conditions=conditions.conditional if conditions else None,
            )
        )

    if len(parents) == 1:
        return parents[0]
    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} Multiple parent DS nodes resolved - {[type(x) for x in parents]}, wrapping in merge"
    )

    preexisting_conditions = None
    if conditions and all(
        [
            x.preexisting_conditions
            and x.preexisting_conditions == conditions.conditional
            for x in parents
        ]
    ):
        preexisting_conditions = conditions.conditional

    base = MergeNode(
        output_concepts=all_concepts,
        input_concepts=non_constant,
        environment=environment,
        depth=depth,
        parents=parents,
        preexisting_conditions=preexisting_conditions,
    )

    return base



from typing import List

from trilogy.constants import logger
from trilogy.core.enums import Derivation
from trilogy.core.exceptions import UnresolvableQueryException
from trilogy.core.models.author import MultiSelectLineage, SelectLineage
from trilogy.core.models.build import (
    BuildConcept,
    BuildGrain,
    BuildRowsetItem,
    BuildRowsetLineage,
    BuildWhereClause,
    Factory,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.processing.nodes import History, MergeNode, StrategyNode
from trilogy.core.processing.utility import concept_to_relevant_joins, padding

LOGGER_PREFIX = "[GEN_ROWSET_NODE]"


def gen_rowset_node(
    concept: BuildConcept,
    local_optional: List[BuildConcept],
    environment: BuildEnvironment,
    g,
    depth: int,
    source_concepts,
    history: History,
    conditions: BuildWhereClause | None = None,
) -> StrategyNode | None:
    from trilogy.core.query_processor import get_query_node

    if not isinstance(concept.lineage, BuildRowsetItem):
        raise SyntaxError(
            f"Invalid lineage passed into rowset fetch, got {type(concept.lineage)}, expected {BuildRowsetItem}"
        )
    lineage: BuildRowsetItem = concept.lineage
    rowset: BuildRowsetLineage = lineage.rowset
    select: SelectLineage | MultiSelectLineage = lineage.rowset.select

    node = get_query_node(history.base_environment, select)

    if not node:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} Cannot generate parent rowset node for {concept}"
        )
        raise UnresolvableQueryException(
            f"Cannot generate parent select for concept {concept} in rowset {rowset.name}; ensure the rowset is a valid statement."
        )
    enrichment = set([x.address for x in local_optional])

    factory = Factory(environment=history.base_environment, grain=select.grain)
    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} rowset derived concepts are {lineage.rowset.derived_concepts}"
    )
    concept_pool = list(environment.concepts.values()) + list(
        environment.alias_origin_lookup.values()
    )
    rowset_outputs = [
        x.address for x in concept_pool if x.address in lineage.rowset.derived_concepts
    ]
    rowset_relevant: list[BuildConcept] = [
        v for v in concept_pool if v.address in rowset_outputs
    ]

    additional_relevant = [
        factory.build(x) for x in select.output_components if x.address in enrichment
    ]
    # add in other other concepts
    node.set_output_concepts(rowset_relevant + additional_relevant)
    if select.where_clause:
        for item in additional_relevant:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} adding {item} to partial concepts"
            )
            node.partial_concepts.append(item)

    node.grain = BuildGrain.from_concepts(
        [
            x
            for x in node.output_concepts
            if x.address
            not in [
                y
                for y in node.hidden_concepts
                if y in environment.concepts
                and environment.concepts[y].derivation != Derivation.ROWSET
            ]
        ],
    )

    node.rebuild_cache()
    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} final output is {[x.address for x in node.output_concepts]} with grain {node.grain}"
    )
    if not local_optional or all(
        (
            x.address in node.output_concepts
            or (z in x.pseudonyms for z in node.output_concepts)
        )
        and x.address not in node.partial_concepts
        for x in local_optional
    ):
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} no enrichment required for rowset node as all optional {[x.address for x in local_optional]} found or no optional; exiting early."
        )
        return node
    remaining = [
        x
        for x in local_optional
        if x not in node.output_concepts or x in node.partial_concepts
    ]
    possible_joins = concept_to_relevant_joins(
        [x for x in node.output_concepts if x.derivation != Derivation.ROWSET]
    )
    if not possible_joins:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} no possible joins for rowset node to get {[x.address for x in local_optional]}; have {[x.address for x in node.output_concepts]}"
        )
        return node
    if any(x.derivation == Derivation.ROWSET for x in possible_joins):

        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} cannot enrich rowset node with rowset concepts; exiting early"
        )
        return node
    logger.info([x.address for x in possible_joins + local_optional])
    enrich_node: MergeNode = source_concepts(  # this fetches the parent + join keys
        # to then connect to the rest of the query
        mandatory_list=possible_joins + remaining,
        environment=environment,
        g=g,
        depth=depth + 1,
        conditions=conditions,
        history=history,
    )
    if not enrich_node:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} Cannot generate rowset enrichment node for {concept} with optional {local_optional}, returning just rowset node"
        )
        return node

    non_hidden = [
        x for x in node.output_concepts if x.address not in node.hidden_concepts
    ]
    for x in possible_joins:
        if x.address in node.hidden_concepts:
            node.unhide_output_concepts([x])
    non_hidden_enrich = [
        x
        for x in enrich_node.output_concepts
        if x.address not in enrich_node.hidden_concepts
    ]
    return MergeNode(
        input_concepts=non_hidden + non_hidden_enrich,
        output_concepts=non_hidden + local_optional,
        environment=environment,
        depth=depth,
        parents=[
            node,
            enrich_node,
        ],
        preexisting_conditions=conditions.conditional if conditions else None,
    )



from collections import defaultdict
from math import ceil
from typing import Dict, List, Optional, Set, Tuple, Union

from trilogy.constants import CONFIG, logger
from trilogy.core.constants import CONSTANT_DATASET
from trilogy.core.enums import BooleanOperator, SourceType
from trilogy.core.env_processor import generate_graph
from trilogy.core.ergonomics import generate_cte_names
from trilogy.core.models.author import MultiSelectLineage, SelectLineage
from trilogy.core.models.build import (
    BuildConcept,
    BuildConditional,
    BuildDatasource,
    BuildFunction,
    BuildMultiSelectLineage,
    BuildParamaterizedConceptReference,
    BuildSelectLineage,
    Factory,
)
from trilogy.core.models.environment import Environment
from trilogy.core.models.execute import (
    CTE,
    BaseJoin,
    CTEConceptPair,
    InstantiatedUnnestJoin,
    Join,
    QueryDatasource,
    RecursiveCTE,
    UnionCTE,
    UnnestJoin,
)
from trilogy.core.optimization import optimize_ctes
from trilogy.core.processing.concept_strategies_v3 import source_query_concepts
from trilogy.core.processing.nodes import History, SelectNode, StrategyNode
from trilogy.core.statements.author import (
    ConceptDeclarationStatement,
    CopyStatement,
    MultiSelectStatement,
    PersistStatement,
    SelectStatement,
)
from trilogy.core.statements.execute import (
    MaterializedDataset,
    ProcessedCopyStatement,
    ProcessedQuery,
    ProcessedQueryPersist,
)
from trilogy.hooks.base_hook import BaseHook
from trilogy.utility import unique

LOGGER_PREFIX = "[QUERY BUILD]"


def base_join_to_join(
    base_join: BaseJoin | UnnestJoin, ctes: List[CTE | UnionCTE]
) -> Join | InstantiatedUnnestJoin:
    """This function converts joins at the datasource level
    to joins at the CTE level"""
    if isinstance(base_join, UnnestJoin):
        object_to_unnest = base_join.parent.arguments[0]
        if not isinstance(
            object_to_unnest,
            (BuildConcept | BuildParamaterizedConceptReference | BuildFunction),
        ):
            raise ValueError(f"Unnest join must be a concept; got {object_to_unnest}")
        return InstantiatedUnnestJoin(
            object_to_unnest=object_to_unnest,
            alias=base_join.alias,
        )

    def get_datasource_cte(
        datasource: BuildDatasource | QueryDatasource,
    ) -> CTE | UnionCTE:
        eligible = set()
        for cte in ctes:
            if cte.source.identifier == datasource.identifier:
                return cte
            eligible.add(cte.source.identifier)
        for cte in ctes:
            if cte.source.datasources[0].identifier == datasource.identifier:
                return cte
            eligible.add(cte.source.datasources[0].identifier)
        raise ValueError(
            f"Could not find CTE for datasource {datasource.identifier}; have {eligible}"
        )

    if base_join.left_datasource is not None:
        left_cte = get_datasource_cte(base_join.left_datasource)
    else:
        # multiple left ctes
        left_cte = None
    right_cte = get_datasource_cte(base_join.right_datasource)
    if base_join.concept_pairs:
        final_pairs = [
            CTEConceptPair(
                left=pair.left,
                right=pair.right,
                existing_datasource=pair.existing_datasource,
                modifiers=pair.modifiers,
                cte=get_datasource_cte(pair.existing_datasource),
            )
            for pair in base_join.concept_pairs
        ]
    elif base_join.concepts and base_join.left_datasource:
        final_pairs = [
            CTEConceptPair(
                left=concept,
                right=concept,
                existing_datasource=base_join.left_datasource,
                modifiers=[],
                cte=get_datasource_cte(
                    base_join.left_datasource,
                ),
            )
            for concept in base_join.concepts
        ]
    else:
        final_pairs = []
    return Join(
        left_cte=left_cte,
        right_cte=right_cte,
        jointype=base_join.join_type,
        joinkey_pairs=final_pairs,
    )


def generate_source_map(
    query_datasource: QueryDatasource, all_new_ctes: List[CTE | UnionCTE]
) -> Tuple[Dict[str, list[str]], Dict[str, list[str]]]:
    source_map: Dict[str, list[str]] = defaultdict(list)
    # now populate anything derived in this level
    for qdk, qdv in query_datasource.source_map.items():
        unnest = [x for x in qdv if isinstance(x, UnnestJoin)]
        for _ in unnest:
            source_map[qdk] = []
        if (
            qdk not in source_map
            and len(qdv) == 1
            and isinstance(list(qdv)[0], UnnestJoin)
        ):
            source_map[qdk] = []
        basic = [x for x in qdv if isinstance(x, BuildDatasource)]
        for base in basic:
            source_map[qdk].append(base.safe_identifier)

        ctes = [x for x in qdv if isinstance(x, QueryDatasource)]
        if ctes:
            names = set([x.safe_identifier for x in ctes])
            matches = [
                cte for cte in all_new_ctes if cte.source.safe_identifier in names
            ]

            if not matches and names:
                raise SyntaxError(
                    f"Missing parent CTEs for source map; expecting {names}, have {[cte.source.safe_identifier for cte in all_new_ctes]}"
                )
            for cte in matches:
                output_address = [
                    x.address
                    for x in cte.output_columns
                    if x.address not in [z.address for z in cte.partial_concepts]
                ]
                if qdk in output_address:
                    source_map[qdk].append(cte.safe_identifier)
            # now do a pass that accepts partials
            for cte in matches:
                if qdk not in source_map:
                    source_map[qdk] = [cte.safe_identifier]
        if qdk not in source_map:
            if not qdv:
                source_map[qdk] = []
            elif CONFIG.validate_missing:
                raise ValueError(
                    f"Missing {qdk} in {source_map}, source map {query_datasource.source_map} "
                )

    # existence lookups use a separate map
    # as they cannot be referenced in row resolution
    existence_source_map: Dict[str, list[str]] = defaultdict(list)
    for ek, ev in query_datasource.existence_source_map.items():
        ids = set([x.safe_identifier for x in ev])
        ematches = [
            cte.name for cte in all_new_ctes if cte.source.safe_identifier in ids
        ]
        existence_source_map[ek] = ematches
    return {
        k: [] if not v else list(set(v)) for k, v in source_map.items()
    }, existence_source_map


def datasource_to_query_datasource(datasource: BuildDatasource) -> QueryDatasource:
    sub_select: Dict[str, Set[Union[BuildDatasource, QueryDatasource, UnnestJoin]]] = {
        **{c.address: {datasource} for c in datasource.concepts},
    }
    concepts = [c for c in datasource.concepts]
    concepts = unique(concepts, "address")
    return QueryDatasource(
        output_concepts=concepts,
        input_concepts=concepts,
        source_map=sub_select,
        grain=datasource.grain,
        datasources=[datasource],
        joins=[],
        partial_concepts=[x.concept for x in datasource.columns if not x.is_complete],
    )


def generate_cte_name(full_name: str, name_map: dict[str, str]) -> str:
    cte_names = generate_cte_names()
    if CONFIG.human_identifiers:
        if full_name in name_map:
            return name_map[full_name]
        suffix = ""
        idx = len(name_map)
        if idx >= len(cte_names):
            int = ceil(idx / len(cte_names))
            suffix = f"_{int}"
        valid = [x for x in cte_names if x + suffix not in name_map.values()]
        lookup = valid[0]
        new_name = f"{lookup}{suffix}"
        name_map[full_name] = new_name
        return new_name
    else:
        return full_name.replace("<", "").replace(">", "").replace(",", "_")


def resolve_cte_base_name_and_alias_v2(
    name: str,
    source: QueryDatasource,
    source_map: Dict[str, list[str]],
    raw_joins: List[Join | InstantiatedUnnestJoin],
) -> Tuple[str | None, str | None]:
    if not source.datasources:
        return None, None
    if (
        isinstance(source.datasources[0], BuildDatasource)
        and not source.datasources[0].name == CONSTANT_DATASET
    ):
        ds = source.datasources[0]
        return ds.safe_location, ds.safe_identifier

    joins: List[Join] = [join for join in raw_joins if isinstance(join, Join)]
    if joins and len(joins) > 0:
        candidates = [x.left_cte.name for x in joins if x.left_cte]
        for join in joins:
            if join.joinkey_pairs:
                candidates += [x.cte.name for x in join.joinkey_pairs if x.cte]
        disallowed = [x.right_cte.name for x in joins]
        try:
            cte = [y for y in candidates if y not in disallowed][0]
            return cte, cte
        except IndexError:
            raise SyntaxError(
                f"Invalid join configuration {candidates} {disallowed} for {name}",
            )
    counts: dict[str, int] = defaultdict(lambda: 0)
    output_addresses = [x.address for x in source.output_concepts]
    input_address = [x.address for x in source.input_concepts]
    for k, v in source_map.items():
        for vx in v:
            if k in output_addresses:
                counts[vx] = counts[vx] + 1

            if k in input_address:
                counts[vx] = counts[vx] + 1

            counts[vx] = counts[vx]
    if counts:
        return max(counts, key=counts.get), max(counts, key=counts.get)  # type: ignore
    return None, None


def datasource_to_cte(
    query_datasource: QueryDatasource, name_map: dict[str, str]
) -> CTE | UnionCTE:
    parents: list[CTE | UnionCTE] = []
    if query_datasource.source_type == SourceType.UNION:
        direct_parents: list[CTE | UnionCTE] = []
        for child in query_datasource.datasources:
            assert isinstance(child, QueryDatasource)
            child_cte = datasource_to_cte(child, name_map=name_map)
            direct_parents.append(child_cte)
            parents += child_cte.parent_ctes
        human_id = generate_cte_name(query_datasource.identifier, name_map)
        final = UnionCTE(
            name=human_id,
            source=query_datasource,
            parent_ctes=parents,
            internal_ctes=direct_parents,
            output_columns=[
                c.with_grain(query_datasource.grain)
                for c in query_datasource.output_concepts
            ],
            grain=direct_parents[0].grain,
            order_by=query_datasource.ordering,
        )
        return final

    if len(query_datasource.datasources) > 1 or any(
        [isinstance(x, QueryDatasource) for x in query_datasource.datasources]
    ):
        all_new_ctes: List[CTE | UnionCTE] = []
        for datasource in query_datasource.datasources:
            if isinstance(datasource, QueryDatasource):
                sub_datasource = datasource
            else:
                sub_datasource = datasource_to_query_datasource(datasource)

            sub_cte = datasource_to_cte(sub_datasource, name_map)
            parents.append(sub_cte)
            all_new_ctes.append(sub_cte)
        source_map, existence_map = generate_source_map(query_datasource, all_new_ctes)

    else:
        # source is the first datasource of the query datasource
        if query_datasource.datasources:

            source = query_datasource.datasources[0]
            # this is required to ensure that constant datasets
            # render properly on initial access; since they have
            # no actual source
            if source.name == CONSTANT_DATASET:
                source_map = {k: [] for k in query_datasource.source_map}
                existence_map = source_map
            else:
                source_map = {
                    k: [] if not v else [source.safe_identifier]
                    for k, v in query_datasource.source_map.items()
                }
                existence_map = source_map
        else:
            source_map = {k: [] for k in query_datasource.source_map}
            existence_map = source_map

    human_id = generate_cte_name(query_datasource.identifier, name_map)

    final_joins = [
        base_join_to_join(join, [x for x in parents if isinstance(x, (CTE, UnionCTE))])
        for join in query_datasource.joins
    ]

    base_name, base_alias = resolve_cte_base_name_and_alias_v2(
        human_id, query_datasource, source_map, final_joins
    )
    cte_class = CTE

    if query_datasource.source_type == SourceType.RECURSIVE:
        cte_class = RecursiveCTE
        # extra_kwargs['left_recursive_concept'] = query_datasource.left
    cte = cte_class(
        name=human_id,
        source=query_datasource,
        # output columns are what are selected/grouped by
        output_columns=[
            c.with_grain(query_datasource.grain)
            for c in query_datasource.output_concepts
        ],
        source_map=source_map,
        existence_source_map=existence_map,
        # related columns include all referenced columns, such as filtering
        joins=final_joins,
        grain=query_datasource.grain,
        group_to_grain=query_datasource.group_required,
        # we restrict parent_ctes to one level
        # as this set is used as the base for rendering the query
        parent_ctes=parents,
        condition=query_datasource.condition,
        partial_concepts=query_datasource.partial_concepts,
        nullable_concepts=query_datasource.nullable_concepts,
        join_derived_concepts=query_datasource.join_derived_concepts,
        hidden_concepts=query_datasource.hidden_concepts,
        base_name_override=base_name,
        base_alias_override=base_alias,
        order_by=query_datasource.ordering,
    )
    if cte.grain != query_datasource.grain:
        raise ValueError("Grain was corrupted in CTE generation")
    for x in cte.output_columns:
        if (
            x.address not in cte.source_map
            and not any(y in cte.source_map for y in x.pseudonyms)
            and CONFIG.validate_missing
        ):
            raise ValueError(
                f"Missing {x.address} in {cte.source_map}, source map {cte.source.source_map.keys()} "
            )

    return cte


def get_query_node(
    environment: Environment,
    statement: SelectLineage | MultiSelectLineage,
    history: History | None = None,
) -> StrategyNode:
    if not statement.output_components:
        raise ValueError(f"Statement has no output components {statement}")
    history = history or History(base_environment=environment)
    logger.info(
        f"{LOGGER_PREFIX} building query node for {statement.output_components} grain {statement.grain}"
    )
    build_statement: BuildSelectLineage | BuildMultiSelectLineage = Factory(
        environment=environment,
    ).build(statement)

    build_environment = environment.materialize_for_select(
        build_statement.local_concepts
    )
    graph = generate_graph(build_environment)

    logger.info(
        f"{LOGGER_PREFIX} getting source datasource for outputs {build_statement.output_components} grain {build_statement.grain}"
    )

    search_concepts: list[BuildConcept] = build_statement.output_components

    ods: StrategyNode = source_query_concepts(
        output_concepts=search_concepts,
        environment=build_environment,
        g=graph,
        conditions=build_statement.where_clause,
        history=history,
    )
    if not ods:
        raise ValueError(
            f"Could not find source query concepts for {[x.address for x in search_concepts]}"
        )
    ds: StrategyNode = ods
    if build_statement.having_clause:
        final = build_statement.having_clause.conditional
        if ds.conditions:
            final = BuildConditional(
                left=ds.conditions,
                right=build_statement.having_clause.conditional,
                operator=BooleanOperator.AND,
            )
        ds = SelectNode(
            output_concepts=build_statement.output_components,
            input_concepts=ds.usable_outputs,
            parents=[ds],
            environment=ds.environment,
            partial_concepts=ds.partial_concepts,
            conditions=final,
        )
    ds.hidden_concepts = build_statement.hidden_components
    ds.ordering = build_statement.order_by
    # TODO: avoid this
    ds.rebuild_cache()
    return ds


def get_query_datasources(
    environment: Environment,
    statement: SelectStatement | MultiSelectStatement,
    hooks: Optional[List[BaseHook]] = None,
) -> QueryDatasource:
    ds = get_query_node(environment, statement.as_lineage(environment))

    final_qds = ds.resolve()

    if hooks:
        for hook in hooks:
            hook.process_root_strategy_node(ds)

    return final_qds


def flatten_ctes(input: CTE | UnionCTE) -> list[CTE | UnionCTE]:
    output = [input]
    for cte in input.parent_ctes:
        output += flatten_ctes(cte)
    return output


def process_auto(
    environment: Environment,
    statement: PersistStatement | SelectStatement,
    hooks: List[BaseHook] | None = None,
):
    if isinstance(statement, PersistStatement):
        return process_persist(environment, statement, hooks)
    elif isinstance(statement, SelectStatement):
        return process_query(environment, statement, hooks)
    elif isinstance(statement, ConceptDeclarationStatement):
        return None
    raise ValueError(f"Do not know how to process {type(statement)}")


def process_persist(
    environment: Environment,
    statement: PersistStatement,
    hooks: List[BaseHook] | None = None,
) -> ProcessedQueryPersist:
    select = process_query(
        environment=environment, statement=statement.select, hooks=hooks
    )

    # build our object to return
    arg_dict = {k: v for k, v in select.__dict__.items()}
    return ProcessedQueryPersist(
        **arg_dict,
        output_to=MaterializedDataset(address=statement.address),
        datasource=statement.datasource,
    )


def process_copy(
    environment: Environment,
    statement: CopyStatement,
    hooks: List[BaseHook] | None = None,
) -> ProcessedCopyStatement:
    select = process_query(
        environment=environment, statement=statement.select, hooks=hooks
    )

    # build our object to return
    arg_dict = {k: v for k, v in select.__dict__.items()}
    return ProcessedCopyStatement(
        **arg_dict,
        target=statement.target,
        target_type=statement.target_type,
    )


def process_query(
    environment: Environment,
    statement: SelectStatement | MultiSelectStatement,
    hooks: List[BaseHook] | None = None,
) -> ProcessedQuery:
    hooks = hooks or []

    root_datasource = get_query_datasources(
        environment=environment, statement=statement, hooks=hooks
    )
    for hook in hooks:
        hook.process_root_datasource(root_datasource)
    # this should always return 1 - TODO, refactor
    root_cte = datasource_to_cte(root_datasource, environment.cte_name_map)

    for hook in hooks:
        hook.process_root_cte(root_cte)
    raw_ctes: List[CTE | UnionCTE] = list(reversed(flatten_ctes(root_cte)))
    seen = dict()
    # we can have duplicate CTEs at this point
    # so merge them together
    for cte in raw_ctes:
        if cte.name not in seen:
            seen[cte.name] = cte
        else:
            # merge them up
            seen[cte.name] = seen[cte.name] + cte
    for cte in raw_ctes:
        cte.parent_ctes = [seen[x.name] for x in cte.parent_ctes]
    deduped_ctes: List[CTE | UnionCTE] = list(seen.values())

    root_cte.limit = statement.limit
    root_cte.hidden_concepts = statement.hidden_components

    final_ctes = optimize_ctes(deduped_ctes, root_cte, statement)

    return ProcessedQuery(
        order_by=root_cte.order_by,
        limit=statement.limit,
        output_columns=statement.output_components,
        ctes=final_ctes,
        base=root_cte,
        hidden_columns=set([x for x in statement.hidden_components]),
        local_concepts=statement.local_concepts,
        locally_derived=statement.locally_derived,
    )



from datetime import date, datetime
from decimal import Decimal
from typing import Any

from trilogy import Environment, Executor
from trilogy.authoring import (
    ArrayType,
    DataType,
    MapType,
    NumericType,
    StructType,
    TraitDataType,
    arg_to_datatype,
)
from trilogy.core.enums import ComparisonOperator, Modifier
from trilogy.core.exceptions import (
    DatasourceColumnBindingData,
    DatasourceColumnBindingError,
    DatasourceModelValidationError,
)
from trilogy.core.models.build import (
    BuildComparison,
    BuildDatasource,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.validation.common import ExpectationType, ValidationTest, easy_query
from trilogy.utility import unique


def type_check(
    input: Any,
    expected_type: (
        DataType | ArrayType | StructType | MapType | NumericType | TraitDataType
    ),
    nullable: bool = True,
) -> bool:
    if input is None and nullable:
        return True

    target_type = expected_type
    while isinstance(target_type, TraitDataType):
        return type_check(input, target_type.data_type, nullable)

    if target_type == DataType.STRING:
        return isinstance(input, str)
    if target_type == DataType.INTEGER:
        return isinstance(input, int)
    if target_type == DataType.BIGINT:
        return isinstance(input, int)  # or check for larger int if needed
    if target_type == DataType.FLOAT or isinstance(target_type, NumericType):
        return (
            isinstance(input, float)
            or isinstance(input, int)
            or isinstance(input, Decimal)
        )
    if target_type == DataType.NUMBER:
        return isinstance(input, (int, float, Decimal))
    if target_type == DataType.NUMERIC:
        return isinstance(input, (int, float, Decimal))
    if target_type == DataType.BOOL:
        return isinstance(input, bool)
    if target_type == DataType.DATE:
        return isinstance(input, date) and not isinstance(input, datetime)
    if target_type == DataType.DATETIME:
        return isinstance(input, datetime)
    if target_type == DataType.TIMESTAMP:
        return isinstance(input, datetime)  # or timestamp type if you have one
    if target_type == DataType.UNIX_SECONDS:
        return isinstance(input, (int, float))  # Unix timestamps are numeric
    if target_type == DataType.DATE_PART:
        return isinstance(
            input, str
        )  # assuming date parts are strings like "year", "month"
    if target_type == DataType.ARRAY or isinstance(target_type, ArrayType):
        return isinstance(input, list)
    if target_type == DataType.MAP or isinstance(target_type, MapType):
        return isinstance(input, dict)
    if target_type == DataType.STRUCT or isinstance(target_type, StructType):
        return isinstance(input, dict)
    if target_type == DataType.NULL:
        return input is None
    if target_type == DataType.UNKNOWN:
        return True
    return False


def validate_datasource(
    datasource: BuildDatasource,
    env: Environment,
    build_env: BuildEnvironment,
    exec: Executor | None = None,
    fix: bool = False,
) -> list[ValidationTest]:
    results: list[ValidationTest] = []
    # we might have merged concepts, where both will map out to the same
    unique_outputs = unique(
        [build_env.concepts[col.concept.address] for col in datasource.columns],
        "address",
    )
    type_query = easy_query(
        concepts=unique_outputs,
        datasource=datasource,
        env=env,
        limit=100,
    )

    rows = []
    if exec:
        type_sql = exec.generate_sql(type_query)[-1]
        try:
            rows = exec.execute_raw_sql(type_sql).fetchall()
        except Exception as e:
            results.append(
                ValidationTest(
                    raw_query=type_query,
                    generated_query=type_sql,
                    check_type=ExpectationType.LOGICAL,
                    expected="valid_sql",
                    result=DatasourceModelValidationError(
                        f"Datasource {datasource.name} failed validation. Error executing type query {type_sql}: {e}"
                    ),
                    ran=True,
                )
            )
            return results
    else:

        results.append(
            ValidationTest(
                raw_query=type_query,
                check_type=ExpectationType.LOGICAL,
                expected="datatype_match",
                result=None,
                ran=False,
            )
        )
        return results
    failures: list[DatasourceColumnBindingData] = []
    cols_with_error = set()
    for row in rows:
        for col in datasource.columns:
            actual_address = build_env.concepts[col.concept.address].safe_address
            if actual_address in cols_with_error:
                continue
            rval = row[actual_address]
            passed = type_check(rval, col.concept.datatype, col.is_nullable)
            if not passed:
                value_type = (
                    arg_to_datatype(rval) if rval is not None else col.concept.datatype
                )
                traits = None
                if isinstance(col.concept.datatype, TraitDataType):
                    traits = col.concept.datatype.traits
                if traits and not isinstance(value_type, TraitDataType):
                    value_type = TraitDataType(type=value_type, traits=traits)
                failures.append(
                    DatasourceColumnBindingData(
                        address=col.concept.address,
                        value=rval,
                        value_type=value_type,
                        value_modifiers=[Modifier.NULLABLE] if rval is None else [],
                        actual_type=col.concept.datatype,
                        actual_modifiers=col.concept.modifiers,
                    )
                )
                cols_with_error.add(actual_address)

    if failures:
        results.append(
            ValidationTest(
                check_type=ExpectationType.LOGICAL,
                expected="datatype_match",
                ran=True,
                result=DatasourceColumnBindingError(
                    address=datasource.identifier, errors=failures
                ),
            )
        )

    query = easy_query(
        concepts=[build_env.concepts[name] for name in datasource.grain.components]
        + [build_env.concepts["grain_check"]],
        datasource=datasource,
        env=exec.environment,
        condition=BuildComparison(
            left=build_env.concepts["grain_check"],
            right=1,
            operator=ComparisonOperator.GT,
        ),
    )
    if not exec:
        results.append(
            ValidationTest(
                raw_query=query,
                check_type=ExpectationType.ROWCOUNT,
                expected="0",
                result=None,
                ran=False,
            )
        )

    else:
        sql = exec.generate_sql(query)[-1]

        rows = exec.execute_raw_sql(sql).fetchmany(10)
        if rows:
            results.append(
                ValidationTest(
                    raw_query=query,
                    generated_query=sql,
                    check_type=ExpectationType.ROWCOUNT,
                    expected="0",
                    result=DatasourceModelValidationError(
                        f"Datasource {datasource.name} failed validation. Found rows that do not conform to grain: {rows}"
                    ),
                    ran=True,
                )
            )

    return results
